from typing import Optional, Tuple

import allel  # type: ignore
import numpy as np
import pandas as pd
import plotly.express as px  # type: ignore
from numpydoc_decorator import doc  # type: ignore

from ..util import CacheMiss, check_types, jitter
from . import base_params, pca_params, plotly_params
from .snp_data import AnophelesSnpData


class AnophelesPca(
    AnophelesSnpData,
):
    def __init__(
        self,
        **kwargs,
    ):
        # N.B., this class is designed to work cooperatively, and
        # so it's important that any remaining parameters are passed
        # to the superclass constructor.
        super().__init__(**kwargs)

    @check_types
    @doc(
        summary="""
            Run a principal components analysis (PCA) using biallelic SNPs from
            the selected genome region and samples.
        """,
        extended_summary="""
            .. versionchanged:: 8.0.0
               SNP ascertainment has changed slightly.

            This function uses biallelic SNPs as input to the PCA. The ascertainment
            of SNPs to include has changed slightly in version 8.0.0 and therefore
            the results of this function may also be slightly different. Previously,
            SNPs were required to be biallelic and one of the observed alleles was
            required to be the reference allele. Now SNPs just have to be biallelic.

            The following additional parameters were also added in version 8.0.0:
            `site_class`, `cohort_size`, `min_cohort_size`, `max_cohort_size`,
            `random_seed`.

        """,
        returns=("df_pca", "evr"),
        notes="""
            This computation may take some time to run, depending on your computing
            environment. Results of this computation will be cached and re-used if
            the `results_cache` parameter was set when instantiating the API client.
        """,
    )
    def pca(
        self,
        region: base_params.regions,
        n_snps: base_params.n_snps,
        n_components: pca_params.n_components = pca_params.n_components_default,
        thin_offset: base_params.thin_offset = 0,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        sample_indices: Optional[base_params.sample_indices] = None,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        site_class: Optional[base_params.site_class] = None,
        min_minor_ac: Optional[
            base_params.min_minor_ac
        ] = pca_params.min_minor_ac_default,
        max_missing_an: Optional[
            base_params.max_missing_an
        ] = pca_params.max_missing_an_default,
        cohort_size: Optional[base_params.cohort_size] = None,
        min_cohort_size: Optional[base_params.min_cohort_size] = None,
        max_cohort_size: Optional[base_params.max_cohort_size] = None,
        exclude_samples: Optional[base_params.samples] = None,
        fit_exclude_samples: Optional[base_params.samples] = None,
        random_seed: base_params.random_seed = 42,
        inline_array: base_params.inline_array = base_params.inline_array_default,
        chunks: base_params.chunks = base_params.native_chunks,
    ) -> Tuple[pca_params.df_pca, pca_params.evr]:
        # Change this name if you ever change the behaviour of this function, to
        # invalidate any previously cached data.
        name = "pca_v4"

        # Normalize params for consistent hash value.
        (
            sample_sets_prepped,
            sample_indices_prepped,
        ) = self._prep_sample_selection_cache_params(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            sample_indices=sample_indices,
        )
        region_prepped = self._prep_region_cache_param(region=region)
        site_mask_prepped = self._prep_optional_site_mask_param(site_mask=site_mask)
        params = dict(
            region=region_prepped,
            n_snps=n_snps,
            thin_offset=thin_offset,
            sample_sets=sample_sets_prepped,
            sample_indices=sample_indices_prepped,
            site_mask=site_mask_prepped,
            site_class=site_class,
            min_minor_ac=min_minor_ac,
            max_missing_an=max_missing_an,
            n_components=n_components,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            exclude_samples=exclude_samples,
            fit_exclude_samples=fit_exclude_samples,
            random_seed=random_seed,
        )

        # Try to retrieve results from the cache.
        try:
            results = self.results_cache_get(name=name, params=params)

        except CacheMiss:
            results = self._pca(chunks=chunks, inline_array=inline_array, **params)
            self.results_cache_set(name=name, params=params, results=results)

        # Unpack results.
        coords = results["coords"]
        evr = results["evr"]
        samples = results["samples"]
        loc_keep_fit = results["loc_keep_fit"]

        # Load sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
        )

        # Ensure aligned with genotype data.
        df_samples = df_samples.set_index("sample_id").loc[samples].reset_index()

        # Combine coords and sample metadata.
        df_coords = pd.DataFrame(
            {f"PC{i + 1}": coords[:, i] for i in range(coords.shape[1])}
        )
        df_pca = df_samples.join(df_coords, how="inner")
        # Add a column for which samples were included in fitting.
        df_pca["pca_fit"] = loc_keep_fit

        return df_pca, evr

    def _pca(
        self,
        *,
        region,
        n_snps,
        thin_offset,
        sample_sets,
        sample_indices,
        site_mask,
        site_class,
        min_minor_ac,
        max_missing_an,
        n_components,
        cohort_size,
        min_cohort_size,
        max_cohort_size,
        exclude_samples,
        fit_exclude_samples,
        random_seed,
        chunks,
        inline_array,
    ):
        # Load diplotypes.
        gn, samples = self.biallelic_diplotypes(
            region=region,
            n_snps=n_snps,
            thin_offset=thin_offset,
            sample_sets=sample_sets,
            sample_indices=sample_indices,
            site_mask=site_mask,
            min_minor_ac=min_minor_ac,
            max_missing_an=max_missing_an,
            site_class=site_class,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
            chunks=chunks,
            inline_array=inline_array,
        )

        with self._spinner(desc="Compute PCA"):
            # Exclude any samples prior to computing PCA.
            if exclude_samples is not None:
                x = np.array(exclude_samples, dtype="U")
                loc_keep = ~np.isin(samples, x)
                samples = samples[loc_keep]
                gn = gn[:, loc_keep]

            # Exclude any samples from fitting only.
            if fit_exclude_samples is not None:
                xf = np.array(fit_exclude_samples, dtype="U")
                loc_keep_fit = ~np.isin(samples, xf)
                gn_fit = gn[:, loc_keep_fit]
            else:
                loc_keep_fit = np.ones(len(samples), dtype=bool)
                gn_fit = gn

            # Remove any sites where all genotypes are identical.
            loc_var = np.any(gn_fit != gn_fit[:, 0, np.newaxis], axis=1)
            gn_fit_var = np.compress(loc_var, gn_fit, axis=0)
            gn_var = np.compress(loc_var, gn, axis=0)

            # Run the PCA.
            if fit_exclude_samples is None:
                # Simple fit and transform on the same data.
                coords, model = allel.pca(gn_fit_var, n_components=n_components)

            else:
                # Fit and transform separately.
                model = allel.stats.decomposition.GenotypePCA(
                    n_components=n_components,
                )
                model.fit(gn_fit_var)
                coords = model.transform(gn_var, copy=False)

            # Work around sign indeterminacy.
            for i in range(coords.shape[1]):
                c = coords[:, i]
                if np.abs(c.min()) > np.abs(c.max()):
                    coords[:, i] = c * -1

        results = dict(
            samples=samples,
            coords=coords,
            evr=model.explained_variance_ratio_,
            loc_keep_fit=loc_keep_fit,
        )
        return results

    @check_types
    @doc(
        summary="""
            Plot explained variance ratios from a principal components analysis
            (PCA) using a plotly bar plot.
        """,
        parameters=dict(
            kwargs="Passed through to px.bar().",
        ),
    )
    def plot_pca_variance(
        self,
        evr: pca_params.evr,
        width: plotly_params.fig_width = 900,
        height: plotly_params.fig_height = 400,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ) -> plotly_params.figure:
        # Prepare plotting variables.
        y = evr * 100  # convert to percent
        x = [str(i + 1) for i in range(len(y))]

        # Set up plotting options.
        plot_kwargs = dict(
            labels={
                "x": "Principal component",
                "y": "Explained variance (%)",
            },
            template="simple_white",
            width=width,
            height=height,
        )
        # Apply any user overrides.
        plot_kwargs.update(kwargs)

        # Make a bar plot.
        fig = px.bar(x=x, y=y, **plot_kwargs)

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig

    @check_types
    @doc(
        summary="""
            Plot sample coordinates from a principal components analysis (PCA)
            as a plotly scatter plot.
        """,
        parameters=dict(
            opacity="Marker opacity.",
            kwargs="Passed through to `px.scatter()`",
        ),
    )
    def plot_pca_coords(
        self,
        data: pca_params.df_pca,
        x: plotly_params.x = "PC1",
        y: plotly_params.y = "PC2",
        color: plotly_params.color = None,
        symbol: plotly_params.symbol = None,
        opacity: float = 0.9,
        jitter_frac: plotly_params.jitter_frac = 0.02,
        random_seed: base_params.random_seed = 42,
        width: plotly_params.fig_width = 900,
        height: plotly_params.fig_height = 600,
        marker_size: plotly_params.marker_size = 10,
        color_discrete_sequence: plotly_params.color_discrete_sequence = None,
        color_discrete_map: plotly_params.color_discrete_map = None,
        category_orders: plotly_params.category_order = None,
        legend_sizing: plotly_params.legend_sizing = "constant",
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        render_mode: plotly_params.render_mode = "svg",
        **kwargs,
    ) -> plotly_params.figure:
        # Copy input data to avoid overwriting.
        data = data.copy()

        # Apply jitter if desired - helps spread out points when tightly clustered.
        if jitter_frac:
            np.random.seed(random_seed)
            data[x] = jitter(data[x], jitter_frac)
            data[y] = jitter(data[y], jitter_frac)

        # Convenience variables.
        data["country_location"] = data["country"] + " - " + data["location"]

        # Normalise color and symbol parameters.
        symbol_prepped = self._setup_sample_symbol(
            data=data,
            symbol=symbol,
        )
        del symbol
        (
            color_prepped,
            color_discrete_map_prepped,
            category_orders_prepped,
        ) = self._setup_sample_colors_plotly(
            data=data,
            color=color,
            color_discrete_map=color_discrete_map,
            color_discrete_sequence=color_discrete_sequence,
            category_orders=category_orders,
        )
        del color
        del color_discrete_map
        del color_discrete_sequence
        del category_orders

        # Configure hover data.
        hover_data = self._setup_sample_hover_data_plotly(
            color=color_prepped, symbol=symbol_prepped
        )

        # Set up plotting options.
        plot_kwargs = dict(
            width=width,
            height=height,
            color=color_prepped,
            symbol=symbol_prepped,
            color_discrete_map=color_discrete_map_prepped,
            category_orders=category_orders_prepped,
            template="simple_white",
            hover_name="sample_id",
            hover_data=hover_data,
            opacity=opacity,
            render_mode=render_mode,
        )

        # Apply any user overrides.
        plot_kwargs.update(kwargs)

        # 2D scatter plot.
        fig = px.scatter(data, x=x, y=y, **plot_kwargs)

        # Tidy up.
        fig.update_layout(
            legend=dict(itemsizing=legend_sizing, tracegroupgap=0),
        )
        fig.update_traces(marker={"size": marker_size})

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig

    @check_types
    @doc(
        summary="""
            Plot sample coordinates from a principal components analysis (PCA)
            as a plotly 3D scatter plot.
        """,
        parameters=dict(
            kwargs="Passed through to `px.scatter_3d()`",
        ),
    )
    def plot_pca_coords_3d(
        self,
        data: pca_params.df_pca,
        x: plotly_params.x = "PC1",
        y: plotly_params.y = "PC2",
        z: plotly_params.z = "PC3",
        color: plotly_params.color = None,
        symbol: plotly_params.symbol = None,
        jitter_frac: plotly_params.jitter_frac = 0.02,
        random_seed: base_params.random_seed = 42,
        width: plotly_params.fig_width = 900,
        height: plotly_params.fig_height = 600,
        marker_size: plotly_params.marker_size = 5,
        color_discrete_sequence: plotly_params.color_discrete_sequence = None,
        color_discrete_map: plotly_params.color_discrete_map = None,
        category_orders: plotly_params.category_order = None,
        legend_sizing: plotly_params.legend_sizing = "constant",
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ) -> plotly_params.figure:
        # Copy input data to avoid overwriting.
        data = data.copy()

        # Apply jitter if desired - helps spread out points when tightly clustered.
        if jitter_frac:
            np.random.seed(random_seed)
            data[x] = jitter(data[x], jitter_frac)
            data[y] = jitter(data[y], jitter_frac)
            data[z] = jitter(data[z], jitter_frac)

        # Convenience variables.
        data["country_location"] = data["country"] + " - " + data["location"]

        # Normalise color and symbol parameters.
        symbol_prepped = self._setup_sample_symbol(
            data=data,
            symbol=symbol,
        )
        del symbol
        (
            color_prepped,
            color_discrete_map_prepped,
            category_orders_prepped,
        ) = self._setup_sample_colors_plotly(
            data=data,
            color=color,
            color_discrete_map=color_discrete_map,
            color_discrete_sequence=color_discrete_sequence,
            category_orders=category_orders,
        )
        del color
        del color_discrete_map
        del color_discrete_sequence
        del category_orders

        # Configure hover data.
        hover_data = self._setup_sample_hover_data_plotly(
            color=color_prepped, symbol=symbol_prepped
        )

        # Set up plotting options.
        plot_kwargs = dict(
            width=width,
            height=height,
            hover_name="sample_id",
            hover_data=hover_data,
            color=color_prepped,
            symbol=symbol_prepped,
            color_discrete_map=color_discrete_map_prepped,
            category_orders=category_orders_prepped,
        )

        # Apply any user overrides.
        plot_kwargs.update(kwargs)

        # 3D scatter plot.
        fig = px.scatter_3d(data, x=x, y=y, z=z, **plot_kwargs)

        # Tidy up.
        fig.update_layout(
            scene=dict(aspectmode="cube"),
            legend=dict(itemsizing=legend_sizing, tracegroupgap=0),
        )
        fig.update_traces(marker={"size": marker_size})

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig



"""Parameters for diplotype clustering functions."""

from typing_extensions import Annotated, TypeAlias, Union, Sequence

from .distance_params import distance_metric
from .clustering_params import linkage_method
from .base_params import transcript


linkage_method_default: linkage_method = "complete"

distance_metric_default: distance_metric = "cityblock"

snp_transcript: TypeAlias = Annotated[
    Union[transcript, Sequence[transcript]],
    "A transcript or a list of transcripts",
]



import numpy as np
import pandas as pd
import plotly.express as px
import scipy.cluster.hierarchy as sch


def plot_dendrogram(
    dist,
    linkage_method,
    count_sort,
    distance_sort,
    render_mode,
    width,
    height,
    title,
    line_width,
    line_color,
    marker_size,
    leaf_data,
    leaf_hover_name,
    leaf_hover_data,
    leaf_color,
    leaf_symbol,
    leaf_y,
    leaf_color_discrete_map,
    leaf_category_orders,
    template,
    y_axis_title,
    y_axis_buffer,
):
    # Hierarchical clustering.
    Z = sch.linkage(dist, method=linkage_method)

    # Compute the dendrogram but don't plot it.
    dend = sch.dendrogram(
        Z,
        count_sort=count_sort,
        distance_sort=distance_sort,
        no_plot=True,
    )

    # Compile the line coordinates into a single dataframe.
    icoord = dend["icoord"]
    dcoord = dend["dcoord"]
    line_segments_x = []
    line_segments_y = []
    for ik, dk in zip(icoord, dcoord):
        # Adding None here breaks up the lines.
        line_segments_x += ik + [None]
        line_segments_y += dk + [None]
    df_line_segments = pd.DataFrame({"x": line_segments_x, "y": line_segments_y})

    # Convert X coordinates to haplotype indices (scipy multiplies coordinates by 10).
    df_line_segments["x"] = (df_line_segments["x"] - 5) / 10

    # Plot the lines.
    fig = px.line(
        df_line_segments,
        x="x",
        y="y",
        render_mode=render_mode,
        template=template,
    )

    # Reorder leaf data to align with dendrogram.
    leaves = dend["leaves"]
    n_leaves = len(leaves)
    leaf_data = leaf_data.iloc[leaves]

    # Add scatter plot to draw the leaves.
    fig.add_traces(
        list(
            px.scatter(
                data_frame=leaf_data,
                x=np.arange(n_leaves),
                y=np.repeat(leaf_y, n_leaves),
                color=leaf_color,
                symbol=leaf_symbol,
                render_mode=render_mode,
                hover_name=leaf_hover_name,
                hover_data=leaf_hover_data,
                template=template,
                color_discrete_map=leaf_color_discrete_map,
                category_orders=leaf_category_orders,
            ).select_traces()
        )
    )

    # Style the lines and markers.
    line_props = dict(
        width=line_width,
        color=line_color,
    )
    marker_props = dict(
        size=marker_size,
    )
    fig.update_traces(line=line_props, marker=marker_props)

    # Style the figure.
    fig.update_layout(
        width=width,
        height=height,
        title=title,
        autosize=True,
        hovermode="closest",
        # I cannot get the xaxis title to appear below the plot, and when
        # it's above the plot it often overlaps the title, so hiding it
        # for now.
        xaxis_title=None,
        yaxis_title=y_axis_title,
        showlegend=True,
    )

    # Style axes.
    fig.update_xaxes(
        mirror=False,
        showgrid=False,
        showline=False,
        showticklabels=False,
        ticks="",
        range=(-2, n_leaves + 2),
    )
    fig.update_yaxes(
        mirror=False,
        showgrid=False,
        showline=False,
        showticklabels=True,
        ticks="outside",
        range=(leaf_y - y_axis_buffer, np.max(dcoord) + y_axis_buffer),
    )

    return fig, leaf_data



from typing import Optional, Tuple

import allel  # type: ignore
import numpy as np
from numpydoc_decorator import doc  # type: ignore
import plotly.graph_objects as go  # type: ignore
import pandas as pd  # type: ignore

from ..util import (
    CacheMiss,
    check_types,
    multiallelic_diplotype_pdist,
    multiallelic_diplotype_mean_sqeuclidean,
    multiallelic_diplotype_mean_cityblock,
)
from ..plotly_dendrogram import plot_dendrogram
from . import (
    base_params,
    plotly_params,
    tree_params,
    clustering_params,
    dipclust_params,
    cnv_params,
)
from .snp_frq import AnophelesSnpFrequencyAnalysis
from .cnv_frq import AnophelesCnvFrequencyAnalysis

AA_CHANGE_QUERY = (
    "effect in ['NON_SYNONYMOUS_CODING', 'START_LOST', 'STOP_LOST', 'STOP_GAINED']"
)


class AnophelesDipClustAnalysis(
    AnophelesCnvFrequencyAnalysis, AnophelesSnpFrequencyAnalysis
):
    def __init__(
        self,
        **kwargs,
    ):
        # N.B., this class is designed to work cooperatively, and
        # so it's important that any remaining parameters are passed
        # to the superclass constructor.
        super().__init__(**kwargs)

    @check_types
    @doc(
        summary=""""
        Hierarchically cluster diplotypes in region and produce an interactive plot.

        If you use this function in a publication, please cite both this package and the original manuscript: doi.org/10.1093/molbev/msae140
        """,
    )
    def plot_diplotype_clustering(
        self,
        region: base_params.regions,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        cohort_size: Optional[base_params.cohort_size] = None,
        random_seed: base_params.random_seed = 42,
        color: plotly_params.color = None,
        symbol: plotly_params.symbol = None,
        linkage_method: dipclust_params.linkage_method = dipclust_params.linkage_method_default,
        distance_metric: dipclust_params.distance_metric = dipclust_params.distance_metric_default,
        count_sort: Optional[tree_params.count_sort] = None,
        distance_sort: Optional[tree_params.distance_sort] = None,
        title: plotly_params.title = True,
        title_font_size: plotly_params.title_font_size = 14,
        width: plotly_params.fig_width = None,
        height: plotly_params.fig_height = 500,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        render_mode: plotly_params.render_mode = "svg",
        leaf_y: clustering_params.leaf_y = 0,
        marker_size: plotly_params.marker_size = 5,
        line_width: plotly_params.line_width = 0.5,
        line_color: plotly_params.line_color = "black",
        color_discrete_sequence: plotly_params.color_discrete_sequence = None,
        color_discrete_map: plotly_params.color_discrete_map = None,
        category_orders: plotly_params.category_order = None,
        legend_sizing: plotly_params.legend_sizing = "constant",
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ) -> Optional[dict]:
        import sys

        # Normalise params.
        if count_sort is None and distance_sort is None:
            count_sort = True
            distance_sort = False

        # This is needed to avoid RecursionError on some clustering analyses
        # with larger numbers of nodes.
        sys.setrecursionlimit(10_000)

        # Load sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        dist, gt_samples, n_snps_used = self.diplotype_pairwise_distances(
            region=region,
            site_mask=site_mask,
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            cohort_size=cohort_size,
            distance_metric=distance_metric,
            random_seed=random_seed,
            chunks=chunks,
            inline_array=inline_array,
        )

        # Align sample metadata with genotypes.
        df_samples = (
            df_samples.set_index("sample_id").loc[gt_samples.tolist()].reset_index()
        )

        # Normalise color and symbol parameters.
        symbol_prepped = self._setup_sample_symbol(
            data=df_samples,
            symbol=symbol,
        )
        del symbol
        (
            color_prepped,
            color_discrete_map_prepped,
            category_orders_prepped,
        ) = self._setup_sample_colors_plotly(
            data=df_samples,
            color=color,
            color_discrete_map=color_discrete_map,
            color_discrete_sequence=color_discrete_sequence,
            category_orders=category_orders,
        )
        del color
        del color_discrete_map
        del color_discrete_sequence

        # Configure hover data.
        hover_data = self._setup_sample_hover_data_plotly(
            color=color_prepped, symbol=symbol_prepped
        )

        # Construct plot title.
        if title is True:
            title_lines = []
            if sample_sets is not None:
                title_lines.append(f"Sample sets: {sample_sets}")
            if sample_query is not None:
                title_lines.append(f"Sample query: {sample_query}")
            title_lines.append(f"Genomic region: {region} ({n_snps_used:,} SNPs)")
            title = "<br>".join(title_lines)

        # Create the plot.
        with self._spinner("Plot dendrogram"):
            fig, leaf_data = plot_dendrogram(
                dist=dist,
                linkage_method=linkage_method,
                count_sort=count_sort,
                distance_sort=distance_sort,
                render_mode=render_mode,
                width=width,
                height=height,
                title=title,
                line_width=line_width,
                line_color=line_color,
                marker_size=marker_size,
                leaf_data=df_samples,
                leaf_hover_name="sample_id",
                leaf_hover_data=hover_data,
                leaf_color=color_prepped,
                leaf_symbol=symbol_prepped,
                leaf_y=leaf_y,
                leaf_color_discrete_map=color_discrete_map_prepped,
                leaf_category_orders=category_orders_prepped,
                template="simple_white",
                y_axis_title=f"Distance ({distance_metric})",
                y_axis_buffer=0.1,
            )

        # Tidy up.
        fig.update_layout(
            title_font=dict(
                size=title_font_size,
            ),
            legend=dict(itemsizing=legend_sizing, tracegroupgap=0),
        )

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return {
                "figure": fig,
                "dendro_sample_id_order": np.asarray(leaf_data["sample_id"].to_list()),
                "n_snps": n_snps_used,
            }

    def diplotype_pairwise_distances(
        self,
        region: base_params.regions,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        site_class: Optional[base_params.site_class] = None,
        cohort_size: Optional[base_params.cohort_size] = None,
        distance_metric: dipclust_params.distance_metric = dipclust_params.distance_metric_default,
        random_seed: base_params.random_seed = 42,
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ) -> Tuple[np.ndarray, np.ndarray, int]:
        # Change this name if you ever change the behaviour of this function, to
        # invalidate any previously cached data.
        name = "diplotype_pairwise_distances_v1"

        # Normalize params for consistent hash value.
        sample_sets_prepped = self._prep_sample_sets_param(sample_sets=sample_sets)
        region_prepped = self._prep_region_cache_param(region=region)
        params = dict(
            region=region_prepped,
            site_mask=site_mask,
            sample_sets=sample_sets_prepped,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            site_class=site_class,
            cohort_size=cohort_size,
            distance_metric=distance_metric,
            random_seed=random_seed,
        )

        # Try to retrieve results from the cache.
        try:
            results = self.results_cache_get(name=name, params=params)

        except CacheMiss:
            results = self._diplotype_pairwise_distances(
                chunks=chunks, inline_array=inline_array, **params
            )
            self.results_cache_set(name=name, params=params, results=results)

        # Unpack results")
        dist: np.ndarray = results["dist"]
        gt_samples: np.ndarray = results["gt_samples"]
        n_snps: int = int(results["n_snps"][()])  # ensure scalar

        return dist, gt_samples, n_snps

    def _diplotype_pairwise_distances(
        self,
        *,
        region,
        site_mask,
        sample_sets,
        sample_query,
        sample_query_options,
        site_class,
        cohort_size,
        distance_metric,
        random_seed,
        chunks,
        inline_array,
    ):
        if distance_metric == "cityblock":
            metric = multiallelic_diplotype_mean_cityblock
        elif distance_metric == "euclidean":
            metric = multiallelic_diplotype_mean_sqeuclidean

        # Load SNP data.
        ds_snps = self.snp_calls(
            region=region,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            site_class=site_class,
            cohort_size=cohort_size,
            random_seed=random_seed,
            chunks=chunks,
            inline_array=inline_array,
        )

        with self._dask_progress(desc="Load genotypes for distance calculation"):
            gt = ds_snps["call_genotype"].data.compute()

        with self._spinner(
            desc="Compute allele counts and remove non-segregating sites"
        ):
            # Compute allele count, remove non-segregating sites.
            ac = allel.GenotypeArray(gt).count_alleles(max_allele=3)
            gt_seg = gt.compress(ac.is_segregating(), axis=0)
            ac_seg = allel.GenotypeArray(gt_seg).to_allele_counts(max_allele=3)
            X = np.ascontiguousarray(np.swapaxes(ac_seg.values, 0, 1))

        # Compute pairwise distances.
        with self._spinner(desc="Compute pairwise distances"):
            dist = multiallelic_diplotype_pdist(X, metric=metric)

        # Extract IDs of samples. Convert to "U" dtype here
        # to allow these to be saved to the results cache.
        gt_samples = ds_snps["sample_id"].values.astype("U")

        return dict(
            dist=dist,
            gt_samples=gt_samples,
            n_snps=np.array(gt_seg.shape[0]),
        )

    @doc(
        summary="Calculate heterozygosity per sample over a region and plot as a track.",
        parameters=dict(
            dendro_sample_id_order="The order of samples in the clustering dendrogram.",
            color_continuous_scale="The colorscale to use for the plot.",
        ),
    )
    def _dipclust_het_bar_trace(
        self,
        *,
        region: base_params.regions,
        dendro_sample_id_order: np.ndarray,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        site_mask: Optional[base_params.site_mask],
        cohort_size: Optional[base_params.cohort_size],
        random_seed: base_params.random_seed,
        color_continuous_scale: Optional[plotly_params.color_continuous_scale],
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ):
        ds_snps = self.snp_calls(
            region=region,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            cohort_size=cohort_size,
            site_mask=site_mask,
            random_seed=random_seed,
            chunks=chunks,
            inline_array=inline_array,
        )

        # Strictly speaking we are loading the genotypes for the second time here,
        # because they would also have been loaded for the diplotype pairwise
        # distance calculation. But it's probably OK as performance is unlikely
        # to be a problem, with genotypes only being loaded for relatively small
        # regions.
        with self._dask_progress(desc="Load genotypes for heterozygosity calculation"):
            gt = allel.GenotypeDaskArray(ds_snps["call_genotype"].data).compute()
        samples = ds_snps["sample_id"].values.astype("U")

        with self._spinner(desc="Compute heterozygosity"):
            het_per_sample = gt.is_het().sum(axis=0) / gt.is_called().sum(axis=0)

        df_het = pd.DataFrame(
            {"sample_id": samples, "Sample Heterozygosity": het_per_sample}
        ).set_index("sample_id")

        # order according to dendrogram and transpose
        df_het = df_het.loc[dendro_sample_id_order, :].T

        trace = go.Heatmap(
            z=df_het,
            y=["Heterozygosity"],
            colorscale=color_continuous_scale,
            showlegend=False,
            showscale=False,
        )

        return trace

    @doc(
        summary="Plot CNV calls as a track.",
        parameters=dict(
            figures="The plotly figures to add the CNV track to.",
            row_heights="The height of each row in the plot.",
            cnv_region="The region to plot CNV calls for.",
            dendro_sample_id_order="The order of samples in the clustering dendrogram.",
            samples="The samples present in the diplotype clustering dendrogram.",
            color_continuous_scale="The colorscale to use for the plot.",
        ),
    )
    def _dipclust_cnv_bar_trace(
        self,
        *,
        cnv_region: base_params.region,
        dendro_sample_id_order: np.ndarray,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        max_coverage_variance: Optional[cnv_params.max_coverage_variance],
        colorscale: Optional[plotly_params.color_continuous_scale],
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ):
        try:
            # TODO The gene_cnv() method still needs to get migrated to the
            # AnophelesCnvData class, so that it can be found in the class
            # hierarchy. This will still work for the time being though,
            # if called via the Ag3 or Af1 classes.
            ds_cnv = self.gene_cnv(  # type: ignore
                region=cnv_region,
                sample_sets=sample_sets,
                sample_query=sample_query,
                sample_query_options=sample_query_options,
                max_coverage_variance=max_coverage_variance,
                chunks=chunks,
                inline_array=inline_array,
            )

        except ValueError:
            return None, 0  # No cnv data

        # Reindex to match the order of samples to the dendrogram, and fill
        # any missing samples with NaN - xarray will do this for us by default.
        ds_cnv_ordered = ds_cnv.set_index(samples="sample_id").reindex(
            samples=dendro_sample_id_order,
        )

        # Get the copy number data to plot.
        cn_mode = ds_cnv_ordered["CN_mode"].values

        # get labels
        gene_names = ds_cnv["gene_name"].values
        gene_ids = ds_cnv["gene_id"].values
        gene_labels = [
            n if not pd.isna(n) else gene_ids[i] for i, n in enumerate(gene_names)
        ]

        cn_mode_no_nan = np.nan_to_num(cn_mode, nan=-1)

        # Plot the copy number data.
        # N.B., here we have to use go.Heatmap directly rather than
        # px.imshow because the latter fails to incorporate zmin,
        # zmax and colorscale within the trace data, and so these
        # then get lost later when we try to combined into a single
        # figure.
        trace = go.Heatmap(
            z=cn_mode_no_nan,
            y=gene_labels,
            zmin=-1,
            zmax=4,
            colorscale=colorscale,
            showlegend=False,
            showscale=False,
        )

        return trace, ds_cnv.sizes["genes"]

    def _dipclust_snp_trace(
        self,
        *,
        transcript: base_params.transcript,
        snp_query: Optional[base_params.snp_query] = AA_CHANGE_QUERY,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        site_mask: Optional[base_params.site_mask],
        dendro_sample_id_order: np.ndarray,
        snp_filter_min_maf: float,
        snp_colorscale: Optional[plotly_params.color_continuous_scale],
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ):
        # load genotype allele counts at SNP variants for each sample
        df_snps = self.snp_genotype_allele_counts(
            transcript=transcript,
            snp_query=snp_query,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            chunks=chunks,
            inline_array=inline_array,
        )
        df_snps = df_snps.set_index("label")

        # set to diplotype cluster order
        df_snps = df_snps.filter(like="count_").loc[
            :, ["count_" + s for s in dendro_sample_id_order]
        ]

        if snp_filter_min_maf:
            df_snps = df_snps.assign(af=lambda x: x.sum(axis=1) / (x.shape[1] * 2))
            df_snps = df_snps.query("af > @snp_filter_min_maf").drop(columns="af")

        if not df_snps.empty:
            n_snps = len(df_snps)
            snp_trace = go.Heatmap(
                z=df_snps.values,
                y=df_snps.index.to_list(),
                colorscale=snp_colorscale,
                showlegend=False,
                showscale=False,
            )
        else:
            n_snps = 0
            snp_trace = None

        return snp_trace, n_snps

    def _dipclust_concat_subplots(
        self,
        figures,
        width,
        height,
        row_heights,
        region: base_params.regions,
        n_snps: int,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
    ):
        from plotly.subplots import make_subplots  # type: ignore

        title_lines = []
        if sample_sets is not None:
            title_lines.append(f"sample sets: {sample_sets}")
        if sample_query is not None:
            title_lines.append(f"sample query: {sample_query}")
        title_lines.append(f"genomic region: {region} ({n_snps} SNPs)")
        title = "<br>".join(title_lines)

        # make subplots
        fig = make_subplots(
            rows=len(figures),
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.02,
            row_heights=row_heights,
        )

        for i, figure in enumerate(figures):
            if isinstance(figure, go.Figure):
                # This is a figure, access the traces within it.
                for trace in range(len(figure["data"])):
                    fig.append_trace(figure["data"][trace], row=i + 1, col=1)
            else:
                # Assume this is a trace, add directly.
                fig.append_trace(figure, row=i + 1, col=1)

        fig.update_xaxes(visible=False)
        fig.update_layout(
            title=title,
            width=width,
            height=height,
            hovermode="closest",
            plot_bgcolor="white",
        )

        return fig

    def _insert_dipclust_snp_trace(
        self,
        *,
        figures,
        subplot_heights,
        snp_row_height: plotly_params.height = 25,
        transcript: base_params.transcript,
        snp_query: Optional[base_params.snp_query] = AA_CHANGE_QUERY,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        site_mask: Optional[base_params.site_mask],
        dendro_sample_id_order: np.ndarray,
        snp_filter_min_maf: float,
        snp_colorscale: Optional[plotly_params.color_continuous_scale],
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ):
        snp_trace, n_snps_transcript = self._dipclust_snp_trace(
            transcript=transcript,
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            snp_query=snp_query,
            site_mask=site_mask,
            dendro_sample_id_order=dendro_sample_id_order,
            snp_filter_min_maf=snp_filter_min_maf,
            snp_colorscale=snp_colorscale,
            chunks=chunks,
            inline_array=inline_array,
        )

        if snp_trace:
            figures.append(snp_trace)
            subplot_heights.append(snp_row_height * n_snps_transcript)
        else:
            print(
                f"No SNPs were found below {snp_filter_min_maf} allele frequency. Omitting SNP genotype plot."
            )
        return figures, subplot_heights

    @doc(
        summary=""""
        Perform diplotype clustering, annotated with heterozygosity, gene copy number and amino acid variants.

        If you use this function in a publication, please cite both this package and the original manuscript: doi.org/10.1093/molbev/msae140
        """,
        parameters=dict(
            heterozygosity="Plot heterozygosity track.",
            snp_transcript="Plot amino acid variants for these transcripts.",
            cnv_region="Plot gene CNV calls for this region.",
            snp_filter_min_maf="Filter amino acid variants with alternate allele frequency below this threshold.",
        ),
    )
    def plot_diplotype_clustering_advanced(
        self,
        region: base_params.regions,
        heterozygosity: bool = True,
        heterozygosity_colorscale: plotly_params.color_continuous_scale = "Greys",
        snp_transcript: Optional[dipclust_params.snp_transcript] = None,
        snp_colorscale: plotly_params.color_continuous_scale = "Greys",
        snp_filter_min_maf: float = 0.05,
        snp_query: Optional[base_params.snp_query] = AA_CHANGE_QUERY,
        cnv_region: Optional[base_params.regions] = None,
        cnv_colorscale: plotly_params.color_continuous_scale = cnv_params.colorscale_default,
        cnv_max_coverage_variance: cnv_params.max_coverage_variance = 0.2,
        site_mask: Optional[base_params.site_mask] = None,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        random_seed: base_params.random_seed = 42,
        cohort_size: Optional[base_params.cohort_size] = None,
        color: plotly_params.color = None,
        symbol: plotly_params.symbol = None,
        linkage_method: dipclust_params.linkage_method = dipclust_params.linkage_method_default,
        distance_metric: dipclust_params.distance_metric = dipclust_params.distance_metric_default,
        count_sort: Optional[tree_params.count_sort] = None,
        distance_sort: Optional[tree_params.distance_sort] = None,
        title: plotly_params.title = True,
        title_font_size: plotly_params.title_font_size = 14,
        width: plotly_params.fig_width = None,
        dendrogram_height: plotly_params.height = 300,
        heterozygosity_height: plotly_params.height = 25,
        snp_row_height: plotly_params.height = 25,
        cnv_row_height: plotly_params.height = 25,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        render_mode: plotly_params.render_mode = "svg",
        leaf_y: clustering_params.leaf_y = 0,
        marker_size: plotly_params.marker_size = 5,
        line_width: plotly_params.line_width = 0.5,
        line_color: plotly_params.line_color = "black",
        color_discrete_sequence: plotly_params.color_discrete_sequence = None,
        color_discrete_map: plotly_params.color_discrete_map = None,
        category_orders: plotly_params.category_order = None,
        legend_sizing: plotly_params.legend_sizing = "constant",
        chunks: base_params.chunks = base_params.native_chunks,
        inline_array: base_params.inline_array = base_params.inline_array_default,
    ):
        if cohort_size and snp_transcript:
            cohort_size = None
            print(
                "Cohort size is not supported with amino acid heatmap. Overriding cohort size to None."
            )

        res = self.plot_diplotype_clustering(
            region=region,
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            site_mask=site_mask,
            count_sort=count_sort,
            distance_metric=distance_metric,
            cohort_size=cohort_size,
            distance_sort=distance_sort,
            linkage_method=linkage_method,
            color=color,
            symbol=symbol,
            title=title,
            title_font_size=title_font_size,
            width=width,
            height=dendrogram_height,
            show=False,
            renderer=renderer,
            render_mode=render_mode,
            leaf_y=leaf_y,
            marker_size=marker_size,
            line_width=line_width,
            line_color=line_color,
            color_discrete_sequence=color_discrete_sequence,
            color_discrete_map=color_discrete_map,
            category_orders=category_orders,
            legend_sizing=legend_sizing,
            random_seed=random_seed,
            chunks=chunks,
            inline_array=inline_array,
        )

        fig_dendro = res["figure"]
        n_snps_cluster = res["n_snps"]
        dendro_sample_id_order = res["dendro_sample_id_order"]

        figures = [fig_dendro]
        subplot_heights = [dendrogram_height]

        if heterozygosity:
            het_trace = self._dipclust_het_bar_trace(
                region=region,
                dendro_sample_id_order=dendro_sample_id_order,
                sample_sets=sample_sets,
                sample_query=sample_query,
                sample_query_options=sample_query_options,
                cohort_size=cohort_size,
                site_mask=site_mask,
                color_continuous_scale=heterozygosity_colorscale,
                random_seed=random_seed,
                chunks=chunks,
                inline_array=inline_array,
            )
            figures.append(het_trace)
            subplot_heights.append(heterozygosity_height)

        if cnv_region:
            cnv_trace, n_cnv_genes = self._dipclust_cnv_bar_trace(
                cnv_region=cnv_region,
                dendro_sample_id_order=dendro_sample_id_order,
                sample_sets=sample_sets,
                sample_query=sample_query,
                sample_query_options=sample_query_options,
                max_coverage_variance=cnv_max_coverage_variance,
                colorscale=cnv_colorscale,
                chunks=chunks,
                inline_array=inline_array,
            )
            # N.B., sometimes no CNV data may be available, so check to
            # see if the trace is not None.
            if cnv_trace is not None:
                figures.append(cnv_trace)
                subplot_heights.append(cnv_row_height * n_cnv_genes)

        if isinstance(snp_transcript, str):
            figures, subplot_heights = self._insert_dipclust_snp_trace(
                transcript=snp_transcript,
                figures=figures,
                subplot_heights=subplot_heights,
                sample_sets=sample_sets,
                sample_query=sample_query,
                sample_query_options=sample_query_options,
                snp_query=snp_query,
                site_mask=site_mask,
                dendro_sample_id_order=dendro_sample_id_order,
                snp_filter_min_maf=snp_filter_min_maf,
                snp_colorscale=snp_colorscale,
                chunks=chunks,
                inline_array=inline_array,
            )
        elif isinstance(snp_transcript, list):
            for st in snp_transcript:
                figures, subplot_heights = self._insert_dipclust_snp_trace(
                    transcript=st,
                    figures=figures,
                    subplot_heights=subplot_heights,
                    sample_sets=sample_sets,
                    sample_query=sample_query,
                    sample_query_options=sample_query_options,
                    snp_query=snp_query,
                    site_mask=site_mask,
                    dendro_sample_id_order=dendro_sample_id_order,
                    snp_filter_min_maf=snp_filter_min_maf,
                    snp_colorscale=snp_colorscale,
                    chunks=chunks,
                    inline_array=inline_array,
                )

        # Calculate total height based on subplot heights, plus a fixed
        # additional component to allow for title, axes etc.
        height = sum(subplot_heights) + 50
        fig = self._dipclust_concat_subplots(
            figures=figures,
            width=width,
            height=height,
            row_heights=subplot_heights,
            sample_sets=sample_sets,
            sample_query=sample_query,  # Only uses query for title.
            region=region,
            n_snps=n_snps_cluster,
        )

        fig["layout"]["yaxis"]["title"] = f"Distance ({distance_metric})"

        # Tidy up.
        fig.update_layout(
            title_font=dict(
                size=title_font_size,
            ),
            legend=dict(itemsizing=legend_sizing, tracegroupgap=0),
        )

        if show:
            fig.show(renderer=renderer)
            return None
        else:
            return fig



"""Parameters for PCA functions."""

import numpy as np
import pandas as pd
from typing_extensions import Annotated, TypeAlias
from . import base_params

n_components: TypeAlias = Annotated[
    int,
    "Number of components to return.",
]

n_components_default: n_components = 20

df_pca: TypeAlias = Annotated[
    pd.DataFrame,
    """
    A dataframe of projections along principal components, one row per sample. The columns are:
        `sample_id` is the identifier of the sample,
        `partner_sample_id` is the identifier of the sample used by the partners who contributed it,
        `contributor` is the partner who contributed the sample,
        `country` is the country the sample was collected in,
        `location` is the location the sample was collected in,
        `year` is the year the sample was collected,
        `month` is the month the sample was collected,
        `latitude` is the latitude of the location the sample was collected in,
        `longitude` is the longitude of the location the sample was collected in,
        `sex_call` is the sex of the sample,
        `sample_set` is the sample set containing the sample,
        `release` is the release containing the sample,
        `quarter` is the quarter of the year the sample was collected,
        `study_id* is the identifier of the study the sample set containing the sample came from,
        `study_url` is the URL of the study the sample set containing the sample came from,
        `terms_of_use_expiry_date` is the date the terms of use for the sample expire,
        `terms_of_use_url` is the URL of the terms of use for the sample,
        `unrestricted_use` indicates whether the sample can be used without restrictions (e.g., if the terms of use of expired),
        `mean_cov` is mean value of the coverage,
        `median_cov` is the median value of the coverage,
        `modal_cov` is the mode of the coverage,
        `mean_cov_2L` is mean value of the coverage on 2L,
        `median_cov_2L` is the median value of the coverage on 2L,
        `mode_cov_2L` is the mode of the coverage on 2L,
        `mean_cov_2R` is mean value of the coverage on 2R,
        `median_cov_2R` is the median value of the coverage on 2R,
        `mode_cov_2R` is the mode of the coverage on 2R,
        `mean_cov_3L` is mean value of the coverage on 3L,
        `median_cov_3L` is the median value of the coverage on 3L,
        `mode_cov_3L` is the mode of the coverage on 3L,
        `mean_cov_3R` is mean value of the coverage on 3R,
        `median_cov_3R` is the median value of the coverage on 3R,
        `mode_cov_3R` is the mode of the coverage on 3R,
        `mean_cov_X` is mean value of the coverage on X,
        `median_cov_X` is the median value of the coverage on X,
        `mode_cov_X` is the mode of the coverage on X,
        `frac_gen_cov` is the faction of the genome covered,
        `divergence` is the divergence,
        `contam_pct` is the percentage of contamination,
        `contam_LLR` is the log-likelihood ratio of contamination,
        `aim_species_fraction_arab` is the fraction of the gambcolu vs. arabiensis AIMs that indicated arabiensis (this column is only present for *Ag3*),
        `aim_species_fraction_colu` is the fraction of the gambiae vs. coluzzii AIMs that indicated coluzzii (this column is only present for *Ag3*),
        `aim_species_fraction_colu_no2l` is the fraction of the gambiae vs. coluzzii AIMs that indicated coluzzii, not including the chromosome arm 2L which contains an introgression (this column is only present for *Ag3*),
        `aim_species_gambcolu_arabiensis` is the taxonomic group assigned by the gambcolu vs. arabiensis AIMs (this column is only present for *Ag3*),
        `aim_species_gambiae_coluzzi` is the taxonomic group assigned by the gambiae vs. coluzzii AIMs (this column is only present for *Ag3*),
        `aim_species_gambcolu_arabiensis` is the taxonomic group assigned by the combination of both AIMs analyses (this column is only present for *Ag3*),
        `country_iso` is the ISO code of the country the sample was collected in,
        `admin1_name` is the name of the first administrative level the sample was collected in,
        `admin1_iso` is the ISO code of the first administrative level the sample was collected in,
        `admin2_name` is the name of the second administrative level the sample was collected in,
        `taxon` is the taxon assigned to the sample by the combination of the AIMs analysis and the cohort analysis,
        `cohort_admin1_year` is the cohort the sample belongs to when samples are grouped by first administrative level and year,
        `cohort_admin1_month` is the cohort the sample belongs to when samples are grouped by first administrative level and month,
        `cohort_admin1_quarter` is the cohort the sample belongs to when samples are grouped by first administrative level and quarter,
        `cohort_admin2_year` is the cohort the sample belongs to when samples are grouped by second administrative level and year,
        `cohort_admin2_month` is the cohort the sample belongs to when samples are grouped by second administrative level and month,
        `cohort_admin2_quarter` is the cohort the sample belong to when samples are grouped by second administrative level and quarter.
        `PC?` is the projection along principal component ? (? being an integer between 1 and the number of components). There are as many such columns as components,
        `pca_fit` is whether this sample was used for fitting.
    """,
]

evr: TypeAlias = Annotated[
    np.ndarray,
    "An array of explained variance ratios, one per component.",
]

min_minor_ac_default: base_params.min_minor_ac = 2

max_missing_an_default: base_params.max_missing_an = 0



import io
from itertools import cycle
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import ipyleaflet  # type: ignore
import numpy as np
import pandas as pd
import plotly.express as px  # type: ignore
from numpydoc_decorator import doc  # type: ignore

from ..util import check_types
from . import base_params, map_params, plotly_params
from .base import AnophelesBase


class AnophelesSampleMetadata(AnophelesBase):
    def __init__(
        self,
        cohorts_analysis: Optional[str] = None,
        aim_analysis: Optional[str] = None,
        aim_metadata_dtype: Optional[Mapping[str, Any]] = None,
        taxon_colors: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        # N.B., this class is designed to work cooperatively, and
        # so it's important that any remaining parameters are passed
        # to the superclass constructor.
        super().__init__(**kwargs)

        # If provided, this analysis version will override the
        # default value provided in the release configuration.
        self._cohorts_analysis_override = cohorts_analysis

        # If provided, this analysis version will override the
        # default value provided in the release configuration.
        self._aim_analysis_override = aim_analysis

        # N.B., the expected AIM metadata columns may vary between
        # data resources, and so column names and dtype need to be
        # passed in as parameters.
        self._aim_metadata_columns: Optional[List[str]] = None
        self._aim_metadata_dtype: Dict[str, Any] = dict()
        if isinstance(aim_metadata_dtype, Mapping):
            self._aim_metadata_columns = list(aim_metadata_dtype.keys())
            self._aim_metadata_dtype.update(aim_metadata_dtype)
        self._aim_metadata_dtype["sample_id"] = "object"

        # Set up taxon colors.
        self._taxon_colors = taxon_colors

        # Set up extra metadata.
        self._extra_metadata: List = []

        # Initialize cache attributes.
        self._cache_sample_metadata: Dict = dict()

    def _metadata_paths(
        self,
        *,
        sample_sets: List[str],
        path_template: str,
        aim_analysis: Optional[str] = None,
        cohorts_analysis: Optional[str] = None,
    ) -> Dict[str, str]:
        paths = dict()
        for sample_set in sample_sets:
            release = self.lookup_release(sample_set=sample_set)
            release_path = self._release_to_path(release=release)
            if aim_analysis:
                path = path_template.format(
                    release_path=release_path,
                    sample_set=sample_set,
                    aim_analysis=aim_analysis,
                )
            elif cohorts_analysis:
                path = path_template.format(
                    release_path=release_path,
                    sample_set=sample_set,
                    cohorts_analysis=cohorts_analysis,
                )
            else:
                path = path_template.format(
                    release_path=release_path, sample_set=sample_set
                )
            paths[sample_set] = path
        return paths

    def _parse_metadata_paths(
        self,
        path_template: str,
        parse_metadata_func: Callable[[str, Union[bytes, Exception]], pd.DataFrame],
        sample_sets: Optional[base_params.sample_sets] = None,
        aim_analysis: Optional[str] = None,
        cohorts_analysis: Optional[str] = None,
    ) -> pd.DataFrame:
        # Normalise input parameters.
        sample_sets_prepped = self._prep_sample_sets_param(sample_sets=sample_sets)
        del sample_sets

        # Obtain paths for all files we need to fetch.
        file_paths: Mapping[str, str] = self._metadata_paths(
            sample_sets=sample_sets_prepped,
            path_template=path_template,
            aim_analysis=aim_analysis,
            cohorts_analysis=cohorts_analysis,
        )

        # Fetch all files. N.B., here is an optimisation, this allows us to fetch
        # multiple files concurrently.
        files: Mapping[str, Union[bytes, Exception]] = self.read_files(
            paths=file_paths.values(), on_error="return"
        )

        # Parse files into DataFrames.
        dfs = []
        for sample_set in sample_sets_prepped:
            path = file_paths[sample_set]
            data = files[path]
            df = parse_metadata_func(sample_set, data)
            dfs.append(df)

        # Concatenate all DataFrames.
        df_ret = pd.concat(dfs, axis=0, ignore_index=True)

        return df_ret

    def _parse_general_metadata(
        self, sample_set: str, data: Union[bytes, Exception]
    ) -> pd.DataFrame:
        if isinstance(data, bytes):
            dtype = {
                "sample_id": "object",
                "partner_sample_id": "object",
                "contributor": "object",
                "country": "object",
                "location": "object",
                "year": "int64",
                "month": "int64",
                "latitude": "float64",
                "longitude": "float64",
                "sex_call": "object",
            }
            df = pd.read_csv(io.BytesIO(data), dtype=dtype, na_values="")

            # Ensure all column names are lower case.
            df.columns = [c.lower() for c in df.columns]  # type: ignore

            # Add a couple of columns for convenience.
            df["sample_set"] = sample_set
            release = self.lookup_release(sample_set=sample_set)
            df["release"] = release

            # Derive a quarter column from month.
            df["quarter"] = df.apply(
                lambda row: ((row.month - 1) // 3) + 1 if row.month > 0 else -1,
                axis="columns",
            )

            # Add study columns.
            study_info = self.lookup_study_info(sample_set=sample_set)
            for column in study_info:
                df[column] = study_info[column]

            # Add terms-of-use columns.
            terms_of_use_info = self.lookup_terms_of_use_info(sample_set=sample_set)
            for column in terms_of_use_info:
                df[column] = terms_of_use_info[column]

            return df

        else:
            raise data

    @check_types
    @doc(
        summary="""
            Read general sample metadata for one or more sample sets into a pandas
            DataFrame.
        """,
        returns="A pandas DataFrame, one row per sample.",
    )
    def general_metadata(
        self, sample_sets: Optional[base_params.sample_sets] = None
    ) -> pd.DataFrame:
        return self._parse_metadata_paths(
            path_template="{release_path}/metadata/general/{sample_set}/samples.meta.csv",
            parse_metadata_func=self._parse_general_metadata,
            sample_sets=sample_sets,
        )

    @property
    def _sequence_qc_metadata_dtype(self):
        # Note: tests expect an ordered dictionary.
        # Note: insertion order in dictionary keys is guaranteed since Python 3.7
        # Note: using nullable dtypes (e.g. Int64 instead of int64) to allow missing data.

        dtype = {
            "sample_id": "object",
            "mean_cov": "Float64",
            "median_cov": "Int64",
            "modal_cov": "Int64",
        }

        for contig in sorted(self.config["CONTIGS"]):
            dtype[f"mean_cov_{contig}"] = "Float64"
            dtype[f"median_cov_{contig}"] = "Int64"
            dtype[f"mode_cov_{contig}"] = "Int64"

        dtype.update(
            {
                "frac_gen_cov": "Float64",
                "divergence": "Float64",
                "contam_pct": "Float64",
                "contam_LLR": "Float64",
            }
        )

        return dtype

    def _parse_sequence_qc_metadata(
        self, sample_set: str, data: Union[bytes, Exception]
    ) -> pd.DataFrame:
        if isinstance(data, bytes):
            # Get the dtype of the constant columns.
            dtype = self._sequence_qc_metadata_dtype

            # Read the CSV using the dtype dict.
            df = pd.read_csv(io.BytesIO(data), dtype=dtype, na_values="")

            return df

        elif isinstance(data, FileNotFoundError):
            # Sequence QC metadata are missing for this sample set,
            # so return a blank DataFrame.

            # Copy the sample ids from the general metadata.
            df_general = self.general_metadata(sample_sets=sample_set)
            df = df_general[["sample_id"]].copy()

            # Add the sequence QC columns with appropriate missing values.
            # For each column, set the value to either NA or NaN.
            for c, dtype in self._sequence_qc_metadata_dtype.items():
                if pd.api.types.is_integer_dtype(dtype):
                    # Note: this creates a column with dtype int64.
                    df[c] = -1
                else:
                    # Note: this creates a column with dtype float64.
                    df[c] = np.nan

            # Set the column data types.
            df = df.astype(self._sequence_qc_metadata_dtype)

            return df

        else:
            raise data

    @check_types
    @doc(
        summary="""
            Access sequence QC metadata for one or more sample sets.
        """,
        returns="""A pandas DataFrame, one row per sample. The columns are:
        `sample_id` is the identifier of the sample,
        `partner_sample_id` is the identifier of the sample used by the partners who contributed it,
        `contributor` is the partner who contributed the sample,
        `country` is the country the sample was collected in,
        `location` is the location the sample was collected in,
        `year` is the year the sample was collected,
        `month` is the month the sample was collected,
        `latitude` is the latitude of the location the sample was collected in,
        `longitude` is the longitude of the location the sample was collected in,
        `sex_call` is the sex of the sample,
        `sample_set` is the sample set containing the sample,
        `release` is the release containing the sample,
        `quarter` is the quarter of the year the sample was collected,
        `study_id* is the identifier of the study the sample set containing the sample came from,
        `study_url` is the URL of the study the sample set containing the sample came from,
        `terms_of_use_expiry_date` is the date the terms of use for the sample expire,
        `terms_of_use_url` is the URL of the terms of use for the sample,
        `unrestricted_use` indicates whether the sample can be used without restrictions (e.g., if the terms of use of expired),
        `mean_cov` is mean value of the coverage,
        `median_cov` is the median value of the coverage,
        `modal_cov` is the mode of the coverage,
        `mean_cov_2L` is mean value of the coverage on 2L,
        `median_cov_2L` is the median value of the coverage on 2L,
        `mode_cov_2L` is the mode of the coverage on 2L,
        `mean_cov_2R` is mean value of the coverage on 2R,
        `median_cov_2R` is the median value of the coverage on 2R,
        `mode_cov_2R` is the mode of the coverage on 2R,
        `mean_cov_3L` is mean value of the coverage on 3L,
        `median_cov_3L` is the median value of the coverage on 3L,
        `mode_cov_3L` is the mode of the coverage on 3L,
        `mean_cov_3R` is mean value of the coverage on 3R,
        `median_cov_3R` is the median value of the coverage on 3R,
        `mode_cov_3R` is the mode of the coverage on 3R,
        `mean_cov_X` is mean value of the coverage on X,
        `median_cov_X` is the median value of the coverage on X,
        `mode_cov_X` is the mode of the coverage on X,
        `frac_gen_cov` is the faction of the genome covered,
        `divergence` is the divergence,
        `contam_pct` is the percentage of contamination,
        `contam_LLR` is the log-likelihood ratio of contamination,
        `aim_species_fraction_arab` is the fraction of the gambcolu vs. arabiensis AIMs that indicated arabiensis (this column is only present for *Ag3*),
        `aim_species_fraction_colu` is the fraction of the gambiae vs. coluzzii AIMs that indicated coluzzii (this column is only present for *Ag3*),
        `aim_species_fraction_colu_no2l` is the fraction of the gambiae vs. coluzzii AIMs that indicated coluzzii, not including the chromosome arm 2L which contains an introgression (this column is only present for *Ag3*),
        `aim_species_gambcolu_arabiensis` is the taxonomic group assigned by the gambcolu vs. arabiensis AIMs (this column is only present for *Ag3*),
        `aim_species_gambiae_coluzzi` is the taxonomic group assigned by the gambiae vs. coluzzii AIMs (this column is only present for *Ag3*),
        `aim_species_gambcolu_arabiensis` is the taxonomic group assigned by the combination of both AIMs analyses (this column is only present for *Ag3*),
        `country_iso` is the ISO code of the country the sample was collected in,
        `admin1_name` is the name of the first administrative level the sample was collected in,
        `admin1_iso` is the ISO code of the first administrative level the sample was collected in,
        `admin2_name` is the name of the second administrative level the sample was collected in,
        `taxon` is the taxon assigned to the sample by the combination of the AIMs analysis and the cohort analysis,
        `cohort_admin1_year` is the cohort the sample belongs to when samples are grouped by first administrative level and year,
        `cohort_admin1_month` is the cohort the sample belongs to when samples are grouped by first administrative level and month,
        `cohort_admin1_quarter` is the cohort the sample belongs to when samples are grouped by first administrative level and quarter,
        `cohort_admin2_year` is the cohort the sample belongs to when samples are grouped by second administrative level and year,
        `cohort_admin2_month` is the cohort the sample belongs to when samples are grouped by second administrative level and month,
        `cohort_admin2_quarter` is the cohort the sample belong to when samples are grouped by second administrative level and quarter.
        """,
    )
    def sequence_qc_metadata(
        self, sample_sets: Optional[base_params.sample_sets] = None
    ) -> pd.DataFrame:
        return self._parse_metadata_paths(
            path_template="{release_path}/metadata/curation/{sample_set}/sequence_qc_stats.csv",
            parse_metadata_func=self._parse_sequence_qc_metadata,
            sample_sets=sample_sets,
        )

    @property
    def _cohorts_analysis(self):
        if self._cohorts_analysis_override:
            return self._cohorts_analysis_override
        else:
            # N.B., this will return None if the key is not present in the
            # config.
            return self.config.get("DEFAULT_COHORTS_ANALYSIS")

    @property
    def _cohorts_metadata_columns(self):
        # Handle changes to columns used in different analyses.
        cols = None
        if self._cohorts_analysis:
            if self._cohorts_analysis < "20230223":
                cols = (
                    "country_iso",
                    "admin1_name",
                    "admin1_iso",
                    "admin2_name",
                    "taxon",
                    "cohort_admin1_year",
                    "cohort_admin1_month",
                    "cohort_admin2_year",
                    "cohort_admin2_month",
                )
            # We assume that cohorts analyses from "20230223" onwards always include quarter
            # columns.
            else:
                cols = (
                    "country_iso",
                    "admin1_name",
                    "admin1_iso",
                    "admin2_name",
                    "taxon",
                    "cohort_admin1_year",
                    "cohort_admin1_month",
                    "cohort_admin1_quarter",
                    "cohort_admin2_year",
                    "cohort_admin2_month",
                    "cohort_admin2_quarter",
                )
        return cols

    @property
    def _cohorts_metadata_dtype(self):
        cols = self._cohorts_metadata_columns
        if cols:
            # All columns are string columns.
            dtype = {c: "object" for c in cols}
            dtype["sample_id"] = "object"
            return dtype

    def _parse_cohorts_metadata(
        self, sample_set: str, data: Union[bytes, Exception]
    ) -> pd.DataFrame:
        if isinstance(data, bytes):
            # Parse CSV data.
            dtype = self._cohorts_metadata_dtype
            df = pd.read_csv(io.BytesIO(data), dtype=dtype, na_values="")

            # Ensure all column names are lower case.
            df.columns = [c.lower() for c in df.columns]  # type: ignore

            # Rename some columns for consistent naming.
            df.rename(
                columns={
                    "adm1_iso": "admin1_iso",
                    "adm1_name": "admin1_name",
                    "adm2_name": "admin2_name",
                },
                inplace=True,
            )

            return df

        elif isinstance(data, FileNotFoundError):
            # Cohorts metadata are missing for this sample set, fill with a blank
            # DataFrame.
            df_general = self.general_metadata(sample_sets=sample_set)
            df = df_general[["sample_id"]].copy()
            for c in self._cohorts_metadata_columns:
                df[c] = np.nan
            df = df.astype(self._cohorts_metadata_dtype)
            return df

        else:
            raise data

    def _require_cohorts_analysis(self):
        if not self._cohorts_analysis:
            raise NotImplementedError(
                "Cohorts data not available for this data resource."
            )

    @check_types
    @doc(
        summary="""
            Access cohort membership metadata for one or more sample sets.
        """,
        returns="A pandas DataFrame, one row per sample.",
    )
    def cohorts_metadata(
        self, sample_sets: Optional[base_params.sample_sets] = None
    ) -> pd.DataFrame:
        self._require_cohorts_analysis()

        return self._parse_metadata_paths(
            path_template="{release_path}/metadata/cohorts_{cohorts_analysis}/{sample_set}/samples.cohorts.csv",
            parse_metadata_func=self._parse_cohorts_metadata,
            sample_sets=sample_sets,
            cohorts_analysis=self._cohorts_analysis,
        )

    @property
    def _aim_analysis(self):
        if self._aim_analysis_override:
            return self._aim_analysis_override
        else:
            # N.B., this will return None if the key is not present in the
            # config.
            return self.config.get("DEFAULT_AIM_ANALYSIS")

    def _parse_aim_metadata(
        self, sample_set: str, data: Union[bytes, Exception]
    ) -> pd.DataFrame:
        assert self._aim_metadata_columns is not None
        assert self._aim_metadata_dtype is not None
        if isinstance(data, bytes):
            # Parse CSV data.
            df = pd.read_csv(
                io.BytesIO(data), dtype=self._aim_metadata_dtype, na_values=""
            )

            # Ensure all column names are lower case.
            df.columns = [c.lower() for c in df.columns]  # type: ignore

            return df

        elif isinstance(data, FileNotFoundError):
            # AIM data are missing for this sample set, fill with a blank DataFrame.
            df_general = self.general_metadata(sample_sets=sample_set)
            df = df_general[["sample_id"]].copy()
            for c in self._aim_metadata_columns:
                df[c] = np.nan
            df = df.astype(self._aim_metadata_dtype)
            return df

        else:
            raise data

    def _require_aim_analysis(self):
        if not self._aim_analysis:
            raise NotImplementedError("AIM data not available for this data resource.")

    @check_types
    @doc(
        summary="""
            Access ancestry-informative marker (AIM) metadata for one or more
            sample sets.
        """,
        returns="A pandas DataFrame, one row per sample.",
    )
    def aim_metadata(
        self, sample_sets: Optional[base_params.sample_sets] = None
    ) -> pd.DataFrame:
        self._require_aim_analysis()

        return self._parse_metadata_paths(
            path_template="{release_path}/metadata/species_calls_aim_{aim_analysis}/{sample_set}/samples.species_aim.csv",
            parse_metadata_func=self._parse_aim_metadata,
            sample_sets=sample_sets,
            aim_analysis=self._aim_analysis,
        )

    @check_types
    @doc(
        summary="""
            Add extra sample metadata, e.g., including additional columns
            which you would like to use to query and group samples.
        """,
        parameters=dict(
            data="""
                A data frame with one row per sample. Must include either a
                "sample_id" or "partner_sample_id" column.
            """,
            on="""
                Name of column to use when merging with sample metadata.
            """,
        ),
        notes="""
            The values in the column containing sample identifiers must be
            unique.
        """,
    )
    def add_extra_metadata(self, data: pd.DataFrame, on: str = "sample_id"):
        # Check parameters.
        if not isinstance(data, pd.DataFrame):
            raise TypeError("`data` parameter must be a pandas DataFrame")
        if on not in data.columns:
            raise ValueError(f"dataframe does not contain column {on!r}")
        if on not in {"sample_id", "partner_sample_id"}:
            raise ValueError(
                "`on` parameter must be either 'sample_id' or 'partner_sample_id'"
            )

        # Check for uniqueness.
        if not data[on].is_unique:
            raise ValueError(f"column {on!r} does not have unique values")

        # check there are matching samples.
        df_samples = self.sample_metadata()
        loc_isec = data[on].isin(df_samples[on])
        if not loc_isec.any():
            raise ValueError("no matching samples found")

        # store extra metadata
        self._extra_metadata.append((on, data.copy()))

    @doc(
        summary="Clear any extra metadata previously added",
    )
    def clear_extra_metadata(self):
        self._extra_metadata = []

    @check_types
    @doc(
        summary="Access sample metadata for one or more sample sets.",
        returns="A dataframe of sample metadata, one row per sample.",
    )
    def sample_metadata(
        self,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        sample_indices: Optional[base_params.sample_indices] = None,
    ) -> pd.DataFrame:
        # Extra parameter checks.
        base_params.validate_sample_selection_params(
            sample_query=sample_query, sample_indices=sample_indices
        )

        # Normalise parameters.
        prepped_sample_sets = self._prep_sample_sets_param(sample_sets=sample_sets)
        del sample_sets
        cache_key = tuple(prepped_sample_sets)

        try:
            # Attempt to retrieve from the cache.
            df_samples = self._cache_sample_metadata[cache_key]

        except KeyError:
            with self._spinner(desc="Load sample metadata"):
                ## Build a single DataFrame using all available metadata.

                # Get the general sample metadata.
                # Note: this includes study and terms-of-use info.
                df_samples = self.general_metadata(sample_sets=prepped_sample_sets)

                # Merge with the sequence QC metadata.
                df_sequence_qc = self.sequence_qc_metadata(
                    sample_sets=prepped_sample_sets
                )

                # Note: merging can change column dtypes
                df_samples = df_samples.merge(
                    df_sequence_qc, on="sample_id", sort=False, how="left"
                )

                # If available, merge with the AIM metadata.
                if self._aim_analysis:
                    df_aim = self.aim_metadata(sample_sets=prepped_sample_sets)
                    df_samples = df_samples.merge(
                        df_aim, on="sample_id", sort=False, how="left"
                    )

                # If available, merge with the cohorts metadata.
                if self._cohorts_analysis:
                    df_cohorts = self.cohorts_metadata(sample_sets=prepped_sample_sets)
                    df_samples = df_samples.merge(
                        df_cohorts, on="sample_id", sort=False, how="left"
                    )

            # Store sample metadata in the cache.
            self._cache_sample_metadata[cache_key] = df_samples

        # Add extra metadata.
        for on, data in self._extra_metadata:
            df_samples = df_samples.merge(data, how="left", on=on)

        # For convenience, apply a sample selection.
        if sample_query is not None:
            # Assume a pandas query string.
            sample_query_options = sample_query_options or {}
            df_samples = df_samples.query(sample_query, **sample_query_options)
            df_samples = df_samples.reset_index(drop=True)
        elif sample_indices is not None:
            # Assume it is an indexer.
            df_samples = df_samples.iloc[sample_indices]
            df_samples = df_samples.reset_index(drop=True)

        return df_samples.copy()

    @check_types
    @doc(
        summary="""
            Create a pivot table showing numbers of samples available by space,
            time and taxon.
        """,
        parameters=dict(
            index="Sample metadata columns to use for the pivot table index.",
            columns="Sample metadata columns to use for the pivot table columns.",
        ),
        returns="Pivot table of sample counts. One row per admin2_year cohort. Unless otherwise specified using the `columns` parameters, the samples are grouped according to their taxon and then counted.",
    )
    def count_samples(
        self,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        index: Union[str, Sequence[str]] = (
            "country",
            "admin1_iso",
            "admin1_name",
            "admin2_name",
            "year",
        ),
        columns: Union[str, Sequence[str]] = "taxon",
    ) -> pd.DataFrame:
        # Load sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        # Create pivot table.
        df_pivot = df_samples.pivot_table(
            index=index,
            columns=columns,
            values="sample_id",
            aggfunc="count",
            fill_value=0,
        )

        return df_pivot

    @check_types
    @doc(
        summary="""
            Plot an interactive map showing sampling locations using ipyleaflet.
        """,
        parameters=dict(
            min_samples="""
                Minimum number of samples required to show a marker for a given
                location.
            """,
            count_by="""
                Metadata column to report counts of samples by for each location.
            """,
        ),
        returns="Ipyleaflet map widget.",
    )
    def plot_samples_interactive_map(
        self,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        basemap: Optional[map_params.basemap] = map_params.basemap_default,
        center: map_params.center = map_params.center_default,
        zoom: map_params.zoom = map_params.zoom_default,
        height: map_params.height = map_params.height_default,
        width: map_params.width = map_params.width_default,
        min_samples: int = 1,
        count_by: str = "taxon",
    ) -> ipyleaflet.Map:
        # Normalise height and width to string
        if isinstance(height, int):
            height = f"{height}px"
        if isinstance(width, int):
            width = f"{width}px"

        # Load sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        # Pivot taxa by locations.
        location_composite_key = [
            "country",
            "admin1_iso",
            "admin1_name",
            "admin2_name",
            "location",
            "latitude",
            "longitude",
        ]
        df_pivot = df_samples.pivot_table(
            index=location_composite_key,
            columns=count_by,
            values="sample_id",
            aggfunc="count",
            fill_value=0,
        )

        # Append aggregations to pivot.
        df_location_aggs = df_samples.groupby(location_composite_key).agg(
            {
                "year": lambda x: ", ".join(str(y) for y in sorted(x.unique())),
                "sample_set": lambda x: ", ".join(str(y) for y in sorted(x.unique())),
                "contributor": lambda x: ", ".join(str(y) for y in sorted(x.unique())),
            }
        )
        df_pivot = df_pivot.merge(
            df_location_aggs, on=location_composite_key, validate="one_to_one"
        )

        # Handle basemap.
        basemap_abbrevs = map_params.basemap_abbrevs

        # Determine basemap_provider via basemap
        if isinstance(basemap, str):
            # Interpret string
            # Support case-insensitive basemap abbreviations
            basemap_str = basemap.lower()
            if basemap_str not in basemap_abbrevs:
                raise ValueError(
                    f"Basemap abbreviation not recognised: {basemap_str!r}; try one of {list(basemap_abbrevs.keys())}"
                )
            basemap_provider = basemap_abbrevs[basemap_str]
        elif basemap is None:
            # Default.
            basemap_provider = ipyleaflet.basemaps.Esri.WorldImagery
        else:
            # Expect dict or TileProvider or TileLayer.
            basemap_provider = basemap

        # Create a map.
        samples_map = ipyleaflet.Map(
            center=center,
            zoom=zoom,
            basemap=basemap_provider,
        )
        scale_control = ipyleaflet.ScaleControl(position="bottomleft")
        samples_map.add(scale_control)
        samples_map.layout.height = height
        samples_map.layout.width = width

        # Add markers.
        count_factors = df_samples[count_by].dropna().sort_values().unique()
        for _, row in df_pivot.reset_index().iterrows():
            title = (
                f"Location: {row.location} ({row.latitude:.3f}, {row.longitude:.3f})"
            )
            title += f"\nAdmin level 2: {row.admin2_name}"
            title += f"\nAdmin level 1: {row.admin1_name} ({row.admin1_iso})"
            title += f"\nCountry: {row.country}"
            title += f"\nYears: {row.year}"
            title += f"\nSample sets: {row.sample_set}"
            title += f"\nContributors: {row.contributor}"
            title += "\nNo. specimens: "
            all_n = 0
            for factor in count_factors:
                # Get the number of samples in this taxon
                n = row[factor]
                # Count the number of samples in all taxa
                all_n += n
                if n > 0:
                    title += f"{n} {factor}; "
            # Only show a marker when there are enough samples
            if all_n >= min_samples:
                marker = ipyleaflet.Marker(
                    location=(row.latitude, row.longitude),
                    draggable=False,
                    title=title,
                )
                samples_map.add(marker)

        return samples_map

    @check_types
    @doc(
        summary="""
            Load a data catalog providing URLs for downloading BAM, VCF and Zarr
            files for samples in a given sample set.
        """,
        returns="""One row per sample, the columns are
        `sample_id`, the identifier of the sample,
        `alignments_bam`, the URL of the alignments BAM file,
        `snp_genotypes_vcf`, the URL of the SNP genotypes VCF file,
        `snp_genotypes_zarr`, the URL of the SNP genotypes Zarr file.""",
    )
    def wgs_data_catalog(self, sample_set: base_params.sample_set):
        # Look up release for sample set.
        release = self.lookup_release(sample_set=sample_set)
        release_path = self._release_to_path(release=release)

        # Load data catalog.
        path = f"{self._base_path}/{release_path}/metadata/general/{sample_set}/wgs_snp_data.csv"
        with self._fs.open(path) as f:
            df = pd.read_csv(f, na_values="")

        # Normalise columns.
        df = df[
            [
                "sample_id",
                "alignments_bam",
                "snp_genotypes_vcf",
                "snp_genotypes_zarr",
            ]
        ]

        return df

    @check_types
    @doc(
        summary="""
            Load a data accessions catalog providing ENA run accessions
            for samples in a given sample set.
        """,
        returns="One row per sample, columns provide run accessions.",
    )
    def wgs_run_accessions(self, sample_set: base_params.sample_set):
        # Look up release for sample set.
        release = self.lookup_release(sample_set=sample_set)
        release_path = self._release_to_path(release=release)

        # Load data catalog.
        path = f"{self._base_path}/{release_path}/metadata/general/{sample_set}/wgs_accession_data.csv"
        with self._fs.open(path) as f:
            df = pd.read_csv(f, na_values="")

        # Normalise columns.
        df = df[
            [
                "sample_id",
                "run_ena",
            ]
        ]

        return df

    def _prep_sample_selection_cache_params(
        self,
        *,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        sample_indices: Optional[base_params.sample_indices],
    ) -> Tuple[List[str], Optional[List[int]]]:
        # Normalise sample sets.
        sample_sets = self._prep_sample_sets_param(sample_sets=sample_sets)

        if sample_query is not None:
            # Resolve query to a list of integers for more cache hits - we
            # do this because there are different ways to write the same pandas
            # query, and so it's better to evaluate the query and use a list of
            # integer indices instead.
            df_samples = self.sample_metadata(sample_sets=sample_sets)
            sample_query_options = sample_query_options or {}
            loc_samples = df_samples.eval(sample_query, **sample_query_options).values
            sample_indices = np.nonzero(loc_samples)[0].tolist()

        return sample_sets, sample_indices

    def _results_cache_add_analysis_params(self, params: dict):
        super()._results_cache_add_analysis_params(params)
        params["cohorts_analysis"] = self._cohorts_analysis
        params["aim_analysis"] = self._aim_analysis

    @check_types
    @doc(
        summary="Get the metadata for a specific sample and sample set.",
        returns="The metadata for the specified sample.",
    )
    def lookup_sample(
        self,
        sample: base_params.sample,
        sample_set: Optional[base_params.sample_set] = None,
    ) -> pd.core.series.Series:
        df_samples = self.sample_metadata(sample_sets=sample_set).set_index("sample_id")
        sample_rec = None
        if isinstance(sample, str):
            sample_rec = df_samples.loc[sample]
        else:
            assert isinstance(sample, int)
            sample_rec = df_samples.iloc[sample]
        return sample_rec

    @check_types
    @doc(
        summary="""
            Plot a bar chart showing the number of samples available, grouped by
            some variable such as country or year.
        """,
        parameters=dict(
            x="Name of sample metadata column to plot on the X axis.",
            color="Name of the sample metadata column to color bars by.",
            sort="If True, sort the bars in size order.",
            kwargs="Passed through to px.bar().",
        ),
    )
    def plot_samples_bar(
        self,
        x: str,
        color: Optional[str] = None,
        sort: bool = True,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        template: plotly_params.template = "plotly_white",
        width: plotly_params.fig_width = 800,
        height: plotly_params.fig_height = 600,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ) -> plotly_params.figure:
        # Load sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        # Special handling for plotting by year.
        if x == "year":
            # Remove samples with missing year.
            df_samples = df_samples.query("year > 0")

        # Construct a long-form dataframe to plot.
        if color:
            grouper: Union[str, List[str]] = [x, color]
        else:
            grouper = x
        df_plot = df_samples.groupby(grouper).agg({"sample_id": "count"}).reset_index()

        # Deal with request to sort by bar size.
        if sort:
            df_sort = (
                df_samples.groupby(x)
                .agg({"sample_id": "count"})
                .reset_index()
                .sort_values("sample_id")
            )
            x_order = df_sort[x].values
            category_orders = kwargs.get("category_orders", dict())
            category_orders.setdefault(x, x_order)
            kwargs["category_orders"] = category_orders

        # Make the plot.
        fig = px.bar(
            df_plot,
            x=x,
            y="sample_id",
            color=color,
            template=template,
            width=width,
            height=height,
            **kwargs,
        )

        # Visual styling.
        fig.update_layout(
            xaxis_title=x.capitalize(),
            yaxis_title="No. samples",
        )
        if color:
            fig.update_layout(legend_title=color.capitalize())

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig

    def _setup_sample_symbol(
        self,
        *,
        data,
        symbol,
    ):
        if symbol is None:
            return None

        # Handle the symbol parameter.
        if isinstance(symbol, str):
            if "cohort_" + symbol in data.columns:
                # Convenience to allow things like "admin1_year" instead of "cohort_admin1_year".
                symbol_prepped = "cohort_" + symbol
            else:
                symbol_prepped = symbol
            if symbol_prepped not in data.columns:
                raise ValueError(
                    f"{symbol_prepped!r} is not a known column in the data."
                )

        else:
            # Custom grouping using queries.
            assert isinstance(symbol, Mapping)
            data["symbol"] = ""
            for key, value in symbol.items():
                data.loc[data.query(value).index, "symbol"] = key
            symbol_prepped = "symbol"

        # Handle missing data in a consistent way.
        data[symbol_prepped] = data[symbol_prepped].fillna("<NA>")

        return symbol_prepped

    def _setup_sample_colors_plotly(
        self,
        *,
        data,
        color,
        color_discrete_sequence,
        color_discrete_map,
        category_orders,
    ):
        # Check for no color.
        if color is None:
            # Bail out early.
            return None, None, None

        # Special handling for taxon colors.
        if color == "taxon" and color_discrete_map is None:
            # Special case, default taxon colors and order.
            color_discrete_map = self._taxon_colors

        if isinstance(color, str):
            if "cohort_" + color in data.columns:
                # Convenience to allow things like "admin1_year" instead of "cohort_admin1_year".
                color_prepped = "cohort_" + color
            else:
                color_prepped = color

            if color_prepped not in data.columns:
                raise ValueError(
                    f"{color_prepped!r} is not a known column in the data."
                )

        else:
            # Custom grouping using queries.
            assert isinstance(color, Mapping)
            data["color"] = ""
            for key, value in color.items():
                data.loc[data.query(value).index, "color"] = key
            color_prepped = "color"

        # Finish handling of color parameter.
        del color

        # Handle missing data in a consistent way.
        data[color_prepped] = data[color_prepped].fillna("<NA>")

        # Obtain the values that we will be mapping to colors.
        color_data_unique_values = data[color_prepped].unique()

        # Now set up color choices.
        if color_discrete_map is None:
            # Choose a color palette.
            if color_discrete_sequence is None:
                if len(color_data_unique_values) <= 10:
                    color_discrete_sequence = px.colors.qualitative.Plotly
                else:
                    color_discrete_sequence = px.colors.qualitative.Alphabet

            # Map values to colors.
            color_discrete_map_prepped = {
                v: c
                for v, c in zip(
                    color_data_unique_values, cycle(color_discrete_sequence)
                )
            }

        else:
            color_discrete_map_prepped = color_discrete_map

        # Consistent color for missing data.
        color_discrete_map_prepped["<NA>"] = "black"

        # Finished handling of color map params.
        del color_discrete_map
        del color_discrete_sequence

        # Define category orders.
        if category_orders is None:
            # Default ordering.
            category_orders_prepped = {color_prepped: color_data_unique_values.tolist()}

        else:
            category_orders_prepped = category_orders

        # Finised handling of category orders.
        del category_orders

        return (
            color_prepped,
            color_discrete_map_prepped,
            category_orders_prepped,
        )

    def _setup_sample_hover_data_plotly(
        self,
        *,
        color,
        symbol,
    ):
        hover_data = [
            "sample_id",
            "partner_sample_id",
            "sample_set",
            "taxon",
            "country",
            "admin1_iso",
            "admin1_name",
            "admin2_name",
            "location",
            "year",
            "month",
        ]
        if color and color not in hover_data:
            hover_data.append(color)
        if symbol and symbol not in hover_data:
            hover_data.append(symbol)
        return hover_data

    def _setup_cohort_queries(
        self,
        cohorts: base_params.cohorts,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query],
        sample_query_options: Optional[base_params.sample_query_options],
        cohort_size: Optional[base_params.cohort_size],
        min_cohort_size: Optional[base_params.min_cohort_size],
    ):
        """Convenience function to normalise the `cohorts` paramater to a
        dictionary mapping cohort labels to sample metadata queries."""

        if isinstance(cohorts, dict):
            # User has supplied a custom dictionary mapping cohort identifiers
            # to pandas queries.
            cohort_queries = cohorts

        else:
            assert isinstance(cohorts, str)
            # User has supplied a column in the sample metadata.
            df_samples = self.sample_metadata(
                sample_sets=sample_sets,
                sample_query=sample_query,
                sample_query_options=sample_query_options,
            )

            # Determine column in dataframe - allow abbreviation.
            if "cohort_" + cohorts in df_samples.columns:
                cohorts = "cohort_" + cohorts
            if cohorts not in df_samples.columns:
                raise ValueError(
                    f"{cohorts!r} is not a known column in the sample metadata"
                )

            # Find cohort labels and build queries dictionary.
            cohort_labels = sorted(df_samples[cohorts].dropna().unique())
            cohort_queries = {coh: f"{cohorts} == '{coh}'" for coh in cohort_labels}

        # Handle sample_query parameter.
        if sample_query is not None:
            cohort_queries = {
                cohort_label: f"({cohort_query}) and ({sample_query})"
                for cohort_label, cohort_query in cohort_queries.items()
            }

        # Check cohort sizes, drop any cohorts which are too small.
        cohort_queries_checked = dict()
        for cohort_label, cohort_query in cohort_queries.items():
            df_cohort_samples = self.sample_metadata(
                sample_sets=sample_sets, sample_query=cohort_query
            )
            n_samples = len(df_cohort_samples)
            if min_cohort_size is not None:
                cohort_size = min_cohort_size
            if cohort_size is not None and n_samples < cohort_size:
                print(
                    f"Cohort ({cohort_label}) has insufficient samples ({n_samples}) for requested cohort size ({cohort_size}), dropping."
                )
            else:
                cohort_queries_checked[cohort_label] = cohort_query

        return cohort_queries_checked

    @check_types
    @doc(
        summary="""
            Read data for a specific cohort set, including cohort size,
            country code, taxon, administrative units name, ISO code, geoBoundaries
            shape ID and representative latitude and longitude points.
        """,
        parameters=dict(
            cohort_set="""
                A cohort set name. Accepted values are:
                "admin1_month", "admin1_quarter", "admin1_year",
                "admin2_month", "admin2_quarter", "admin2_year".
            """
        ),
        returns="""A dataframe of cohort data, one row per cohort. There are up to 18 columns:
        `cohort_id` is the identifier of the cohort,
        `cohort_size` is the number of samples in the cohort,
        `country` is the country the cohort is from,
        `country_alpha2` is the ISO alpha-2 code for the country the cohort is from,
        `country_alpha3` is the ISO alpha-3 code for the country the cohort is from,
        `taxon` is the taxon of the samples in the cohort,
        `year` is the year the samples in the cohort were collected in,
        `quarter` is the quarter the samples in the cohort were collected in (this column is only present if the temporal dimension is *quarter* or *month*),
        `month` is the month the samples in the cohort were collected in (this column is only present if the temporal dimension is *month*),
        `admin1_name` is the name of the first administrative level the samples in the cohort were collected in,
        `admin1_iso` is the ISO code of the first administrative level the samples in the cohort were collected in,
        `admin1_geoboundaries_shape_id` is the identifier of the geoboundary shape corresponding to the first administrative level the samples in the cohort were collected in,
        `admin1_representative_longitude` is the representative longitude for the first administrative level the samples in the cohort were collected in,
        `admin1_representative_latitude` is the representative latitude for the first administrative level the samples in the cohort were collected in,
        `admin2_name` is the name of the second administrative level the samples in the cohort were collected in (this column is only present if the spatial dimension is *admin2*),
        `admin2_iso` is the ISO code of the second administrative level the samples in the cohort were collected in (this column is only present if the spatial dimension is *admin2*),
        `admin2_geoboundaries_shape_id` is the identifier of the geoboundary shape corresponding to the second administrative level the samples in the cohort were collected in (this column is only present if the spatial dimension is *admin2*),
        `admin2_representative_longitude` is the representative longitude for the second administrative level the samples in the cohort were collected in (this column is only present if the spatial dimension is *admin2*),
        `admin2_representative_latitude` is the representative latitude for the second administrative level the samples in the cohort were collected in (this column is only present if the spatial dimension is *admin2*).
        """,
    )
    def cohorts(
        self,
        cohort_set: base_params.cohorts,
    ) -> pd.DataFrame:
        major_version_path = self._major_version_path
        cohorts_analysis = self._cohorts_analysis

        path = f"{major_version_path[:2]}_cohorts/cohorts_{cohorts_analysis}/cohorts_{cohort_set}.csv"

        # Read the manifest into a pandas dataframe.
        with self.open_file(path) as f:
            df_cohorts = pd.read_csv(f, sep=",", na_values="")

        # Ensure all column names are lower case.
        df_cohorts.columns = [c.lower() for c in df_cohorts.columns]  # type: ignore

        return df_cohorts

    @check_types
    @doc(
        summary="""
            Plot markers on a map showing sample locations
            as a Mapbox scatter plot.
        """,
        parameters=dict(
            kwargs="Passed through to px.scatter_mapbox().",
        ),
    )
    def plot_sample_location_mapbox(
        self,
        *,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        marker_size: plotly_params.marker_size = 10,
        color: plotly_params.color = "admin1_name",
        color_discrete_sequence: plotly_params.color_discrete_sequence = px.colors.qualitative.Prism,
        category_orders: plotly_params.category_order = None,
        hover_name: plotly_params.hover_name = "location",
        zoom: plotly_params.zoom = None,
        width: plotly_params.fig_width = 800,
        height: plotly_params.fig_height = 600,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ) -> plotly_params.figure:
        # Get the sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        # Set the location columns to use from the sample metadata.
        location_columns = [
            "country",
            "admin1_iso",
            "admin1_name",
            "admin2_name",
            "location",
            "latitude",
            "longitude",
        ]

        # Trim and dedupe the sample locations.
        # Sort by `color` column by default, which can be overridden via category_orders.
        df_locations = df_samples[location_columns].drop_duplicates().sort_values(color)

        fig = px.scatter_mapbox(
            df_locations,
            lat="latitude",
            lon="longitude",
            mapbox_style="open-street-map",
            zoom=zoom,
            color=color,
            category_orders=category_orders,
            color_discrete_sequence=color_discrete_sequence,
            hover_name=hover_name,
            hover_data=location_columns,
            width=width,
            height=height,
            **kwargs,
        )

        # Set the size of the markers.
        fig.update_traces(marker=dict(size=marker_size))

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig

    @check_types
    @doc(
        summary="""
            Plot markers on a map showing sample locations
            as a geographic scatter plot.
        """,
        parameters=dict(
            kwargs="Passed through to px.scatter_mapbox().",
        ),
    )
    def plot_sample_location_geo(
        self,
        *,
        sample_sets: Optional[base_params.sample_sets],
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        marker_size: plotly_params.marker_size = 10,
        color: plotly_params.color = "admin1_name",
        color_discrete_sequence: plotly_params.color_discrete_sequence = px.colors.qualitative.Prism,
        category_orders: plotly_params.category_order = None,
        hover_name: plotly_params.hover_name = "location",
        fitbounds: plotly_params.fitbounds = "locations",
        scope: plotly_params.scope = "world",
        width: plotly_params.fig_width = 800,
        height: plotly_params.fig_height = 600,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ) -> plotly_params.figure:
        # Get the sample metadata.
        df_samples = self.sample_metadata(
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
        )

        # Set the location columns to use from the sample metadata.
        location_columns = [
            "country",
            "admin1_iso",
            "admin1_name",
            "admin2_name",
            "location",
            "latitude",
            "longitude",
        ]

        # Trim and dedupe the sample locations.
        # Sort by `color` column by default, which can be overridden via category_orders.
        df_locations = df_samples[location_columns].drop_duplicates().sort_values(color)

        fig = px.scatter_geo(
            df_locations,
            lat="latitude",
            lon="longitude",
            scope=scope,
            height=height,
            width=width,
            color=color,
            hover_name=hover_name,
            hover_data=location_columns,
            category_orders=category_orders,
            color_discrete_sequence=color_discrete_sequence,
            fitbounds=fitbounds,
            **kwargs,
        )

        # Set the size of the markers.
        fig.update_traces(marker=dict(size=marker_size))

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig


def locate_cohorts(*, cohorts, data, min_cohort_size):
    # Build cohort dictionary where key=cohort_id, value=loc_coh.
    coh_dict = {}

    if isinstance(cohorts, Mapping):
        # User has supplied a custom dictionary mapping cohort identifiers
        # to pandas queries.

        for coh, query in cohorts.items():
            loc_coh = data.eval(query).values
            coh_dict[coh] = loc_coh

    else:
        assert isinstance(cohorts, str)
        # User has supplied the name of a sample metadata column.

        # Convenience to allow things like "admin1_year" instead of "cohort_admin1_year".
        if "cohort_" + cohorts in data.columns:
            cohorts = "cohort_" + cohorts

        # Check the given cohort set exists.
        if cohorts not in data.columns:
            raise ValueError(f"{cohorts!r} is not a known column in the data.")
        cohort_labels = data[cohorts].dropna().unique()

        # Remove the nans and sort.
        cohort_labels = sorted([c for c in cohort_labels if isinstance(c, str)])
        for coh in cohort_labels:
            loc_coh = data[cohorts] == coh
            coh_dict[coh] = loc_coh.values

    # Remove cohorts below minimum cohort size.
    coh_dict = {
        coh: loc_coh
        for coh, loc_coh in coh_dict.items()
        if np.count_nonzero(loc_coh) >= min_cohort_size
    }

    # Early check for no cohorts.
    if len(coh_dict) == 0:
        raise ValueError(
            "No cohorts available for the given sample selection parameters and minimum cohort size."
        )

    return coh_dict



"""Parameters for Plink converter functions."""

from typing_extensions import Annotated, TypeAlias

overwrite: TypeAlias = Annotated[
    bool,
    """
    A boolean indicating whether a previously written file with the same name ought
    to be overwritten. Default is False.
    """,
]

output_dir: TypeAlias = Annotated[
    str,
    """
    A string indicating the desired output file location.
    """,
]



from typing import Tuple, Optional

import numpy as np
import pandas as pd
import allel  # type: ignore
from numpydoc_decorator import doc  # type: ignore
import bokeh.models
import bokeh.plotting
import bokeh.layouts
import plotly.express as px

from .snp_data import AnophelesSnpData
from . import base_params, fst_params, gplt_params, plotly_params
from ..util import CacheMiss, check_types


class AnophelesFstAnalysis(
    AnophelesSnpData,
):
    def __init__(
        self,
        **kwargs,
    ):
        # N.B., this class is designed to work cooperatively, and
        # so it's important that any remaining parameters are passed
        # to the superclass constructor.
        super().__init__(**kwargs)

    def _fst_gwss(
        self,
        *,
        contig,
        window_size,
        sample_sets,
        cohort1_query,
        cohort2_query,
        sample_query_options,
        site_mask,
        cohort_size,
        min_cohort_size,
        max_cohort_size,
        random_seed,
        inline_array,
        chunks,
        clip_min,
    ):
        # Compute allele counts.
        ac1 = self.snp_allele_counts(
            region=contig,
            sample_query=cohort1_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
            inline_array=inline_array,
            chunks=chunks,
        )
        ac2 = self.snp_allele_counts(
            region=contig,
            sample_query=cohort2_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
            inline_array=inline_array,
            chunks=chunks,
        )

        with self._spinner(desc="Load SNP positions"):
            pos = self.snp_sites(
                region=contig,
                field="POS",
                site_mask=site_mask,
                inline_array=inline_array,
                chunks=chunks,
            ).compute()

        with self._spinner(desc="Compute Fst"):
            with np.errstate(divide="ignore", invalid="ignore"):
                fst = allel.moving_hudson_fst(ac1, ac2, size=window_size)
                # Sometimes Fst can be very slightly below zero, clip for simplicity.
                fst = np.clip(fst, a_min=clip_min, a_max=1)
                x = allel.moving_statistic(pos, statistic=np.mean, size=window_size)

        results = dict(x=x, fst=fst)

        return results

    @check_types
    @doc(
        summary="""
            Run a Fst genome-wide scan to investigate genetic differentiation
            between two cohorts.
        """,
        returns=dict(
            x="An array containing the window centre point genomic positions",
            fst="An array with Fst statistic values for each window.",
        ),
    )
    def fst_gwss(
        self,
        contig: base_params.contig,
        window_size: fst_params.window_size,
        cohort1_query: base_params.sample_query,
        cohort2_query: base_params.sample_query,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        sample_sets: Optional[base_params.sample_sets] = None,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        cohort_size: Optional[base_params.cohort_size] = fst_params.cohort_size_default,
        min_cohort_size: Optional[
            base_params.min_cohort_size
        ] = fst_params.min_cohort_size_default,
        max_cohort_size: Optional[
            base_params.max_cohort_size
        ] = fst_params.max_cohort_size_default,
        random_seed: base_params.random_seed = 42,
        inline_array: base_params.inline_array = base_params.inline_array_default,
        chunks: base_params.chunks = base_params.native_chunks,
        clip_min: fst_params.clip_min = 0.0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        # Change this name if you ever change the behaviour of this function, to
        # invalidate any previously cached data.
        name = "fst_gwss_v2"

        params = dict(
            contig=contig,
            window_size=window_size,
            cohort1_query=cohort1_query,
            cohort2_query=cohort2_query,
            sample_query_options=sample_query_options,
            sample_sets=self._prep_sample_sets_param(sample_sets=sample_sets),
            site_mask=self._prep_optional_site_mask_param(site_mask=site_mask),
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
            clip_min=clip_min,
        )

        try:
            results = self.results_cache_get(name=name, params=params)

        except CacheMiss:
            results = self._fst_gwss(**params, inline_array=inline_array, chunks=chunks)
            self.results_cache_set(name=name, params=params, results=results)

        x = results["x"]
        fst = results["fst"]

        return x, fst

    @check_types
    @doc(
        summary="""
            Run and plot a Fst genome-wide scan to investigate genetic
            differentiation between two cohorts.
        """,
    )
    def plot_fst_gwss_track(
        self,
        contig: base_params.contig,
        window_size: fst_params.window_size,
        cohort1_query: base_params.sample_query,
        cohort2_query: base_params.sample_query,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        sample_sets: Optional[base_params.sample_sets] = None,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        cohort_size: Optional[base_params.cohort_size] = fst_params.cohort_size_default,
        min_cohort_size: Optional[
            base_params.min_cohort_size
        ] = fst_params.min_cohort_size_default,
        max_cohort_size: Optional[
            base_params.max_cohort_size
        ] = fst_params.max_cohort_size_default,
        random_seed: base_params.random_seed = 42,
        title: Optional[gplt_params.title] = None,
        sizing_mode: gplt_params.sizing_mode = gplt_params.sizing_mode_default,
        width: gplt_params.width = gplt_params.width_default,
        height: gplt_params.height = 200,
        show: gplt_params.show = True,
        x_range: Optional[gplt_params.x_range] = None,
        output_backend: gplt_params.output_backend = gplt_params.output_backend_default,
        clip_min: fst_params.clip_min = 0.0,
    ) -> gplt_params.optional_figure:
        # compute Fst
        x, fst = self.fst_gwss(
            contig=contig,
            window_size=window_size,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            cohort1_query=cohort1_query,
            cohort2_query=cohort2_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            random_seed=random_seed,
            clip_min=clip_min,
        )

        # determine X axis range
        x_min = x[0]
        x_max = x[-1]
        if x_range is None:
            x_range = bokeh.models.Range1d(x_min, x_max, bounds="auto")

        # create a figure
        xwheel_zoom = bokeh.models.WheelZoomTool(
            dimensions="width", maintain_focus=False
        )
        if title is None:
            title = f"Cohort 1: {cohort1_query}\nCohort 2: {cohort2_query}"
        fig = bokeh.plotting.figure(
            title=title,
            tools=[
                "xpan",
                "xzoom_in",
                "xzoom_out",
                xwheel_zoom,
                "reset",
                "save",
                "crosshair",
            ],
            active_inspect=None,
            active_scroll=xwheel_zoom,
            active_drag="xpan",
            sizing_mode=sizing_mode,
            width=width,
            height=height,
            toolbar_location="above",
            x_range=x_range,
            y_range=(0, 1),
            output_backend=output_backend,
        )

        # plot Fst
        fig.scatter(
            x=x,
            y=fst,
            size=3,
            marker="circle",
            line_width=1,
            line_color="black",
            fill_color=None,
        )

        # tidy up the plot
        fig.yaxis.axis_label = "Fst"
        fig.yaxis.ticker = [0, 1]
        self._bokeh_style_genome_xaxis(fig, contig)

        if show:  # pragma: no cover
            bokeh.plotting.show(fig)
            return None
        else:
            return fig

    @check_types
    @doc(
        summary="""
            Run and plot a Fst genome-wide scan to investigate genetic
            differentiation between two cohorts.
        """,
    )
    def plot_fst_gwss(
        self,
        contig: base_params.contig,
        window_size: fst_params.window_size,
        cohort1_query: base_params.sample_query,
        cohort2_query: base_params.sample_query,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        sample_sets: Optional[base_params.sample_sets] = None,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        cohort_size: Optional[base_params.cohort_size] = fst_params.cohort_size_default,
        min_cohort_size: Optional[
            base_params.min_cohort_size
        ] = fst_params.min_cohort_size_default,
        max_cohort_size: Optional[
            base_params.max_cohort_size
        ] = fst_params.max_cohort_size_default,
        random_seed: base_params.random_seed = 42,
        title: Optional[gplt_params.title] = None,
        sizing_mode: gplt_params.sizing_mode = gplt_params.sizing_mode_default,
        width: gplt_params.width = gplt_params.width_default,
        track_height: gplt_params.track_height = 190,
        genes_height: gplt_params.genes_height = gplt_params.genes_height_default,
        show: gplt_params.show = True,
        output_backend: gplt_params.output_backend = gplt_params.output_backend_default,
        clip_min: fst_params.clip_min = 0.0,
        gene_labels: Optional[gplt_params.gene_labels] = None,
        gene_labelset: Optional[gplt_params.gene_labelset] = None,
    ) -> gplt_params.optional_figure:
        # gwss track
        fig1 = self.plot_fst_gwss_track(
            contig=contig,
            window_size=window_size,
            cohort1_query=cohort1_query,
            cohort2_query=cohort2_query,
            sample_query_options=sample_query_options,
            sample_sets=sample_sets,
            site_mask=site_mask,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
            title=title,
            sizing_mode=sizing_mode,
            width=width,
            height=track_height,
            show=False,
            output_backend=output_backend,
            clip_min=clip_min,
        )

        fig1.xaxis.visible = False

        # plot genes
        fig2 = self.plot_genes(
            region=contig,
            sizing_mode=sizing_mode,
            width=width,
            height=genes_height,
            x_range=fig1.x_range,
            show=False,
            output_backend=output_backend,
            gene_labels=gene_labels,
            gene_labelset=gene_labelset,
        )

        # combine plots into a single figure
        fig = bokeh.layouts.gridplot(
            [fig1, fig2],
            ncols=1,
            toolbar_location="above",
            merge_tools=True,
            sizing_mode=sizing_mode,
            toolbar_options=dict(active_inspect=None),
        )

        if show:  # pragma: no cover
            bokeh.plotting.show(fig)
            return None
        else:
            return fig

    @check_types
    @doc(
        summary="""
            Compute average Hudson's Fst between two specified cohorts.
        """,
        returns="""
            A NumPy float of the Fst value and the standard error (SE).
        """,
    )
    def average_fst(
        self,
        region: base_params.region,
        cohort1_query: base_params.sample_query,
        cohort2_query: base_params.sample_query,
        sample_query_options: Optional[base_params.sample_query] = None,
        sample_sets: Optional[base_params.sample_sets] = None,
        cohort_size: Optional[base_params.cohort_size] = fst_params.cohort_size_default,
        min_cohort_size: Optional[
            base_params.min_cohort_size
        ] = fst_params.min_cohort_size_default,
        max_cohort_size: Optional[
            base_params.max_cohort_size
        ] = fst_params.max_cohort_size_default,
        n_jack: base_params.n_jack = 200,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        site_class: Optional[base_params.site_class] = None,
        random_seed: base_params.random_seed = 42,
    ) -> Tuple[float, float]:
        # Calculate allele counts for each cohort.
        ac1 = self.snp_allele_counts(
            region=region,
            sample_sets=sample_sets,
            sample_query=cohort1_query,
            sample_query_options=sample_query_options,
            cohort_size=cohort_size,
            site_mask=site_mask,
            site_class=site_class,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
        )
        ac2 = self.snp_allele_counts(
            region=region,
            sample_sets=sample_sets,
            sample_query=cohort2_query,
            sample_query_options=sample_query_options,
            cohort_size=cohort_size,
            site_mask=site_mask,
            site_class=site_class,
            min_cohort_size=min_cohort_size,
            max_cohort_size=max_cohort_size,
            random_seed=random_seed,
        )

        # Calculate block length for jackknife.
        n_sites = ac1.shape[0]  # number of sites
        block_length = n_sites // n_jack  # number of sites in each block

        # Calculate average Fst.
        fst, se, _, _ = allel.blockwise_hudson_fst(ac1, ac2, blen=block_length)

        # Normalise to Python scalar types.
        fst = float(fst)
        se = float(se)

        # Fst estimate can sometimes be slightly negative, but clip at
        # zero.
        if fst < 0:
            fst = 0.0

        return fst, se

    @check_types
    @doc(
        summary="""
            Compute pairwise average Hudson's Fst between a set of specified cohorts.
        """,
    )
    def pairwise_average_fst(
        self,
        region: base_params.region,
        cohorts: base_params.cohorts,
        sample_sets: Optional[base_params.sample_sets] = None,
        sample_query: Optional[base_params.sample_query] = None,
        sample_query_options: Optional[base_params.sample_query_options] = None,
        cohort_size: Optional[base_params.cohort_size] = fst_params.cohort_size_default,
        min_cohort_size: Optional[
            base_params.min_cohort_size
        ] = fst_params.min_cohort_size_default,
        max_cohort_size: Optional[
            base_params.max_cohort_size
        ] = fst_params.max_cohort_size_default,
        n_jack: base_params.n_jack = 200,
        site_mask: Optional[base_params.site_mask] = base_params.DEFAULT,
        site_class: Optional[base_params.site_class] = None,
        random_seed: base_params.random_seed = 42,
    ) -> fst_params.df_pairwise_fst:
        # Set up cohort queries.
        cohorts_checked = self._setup_cohort_queries(
            cohorts,
            sample_sets=sample_sets,
            sample_query=sample_query,
            sample_query_options=sample_query_options,
            cohort_size=cohort_size,
            min_cohort_size=min_cohort_size,
        )

        cohort_ids = list(cohorts_checked.keys())
        cohort_queries = list(cohorts_checked.values())
        cohort1_ids = []
        cohort2_ids = []
        fst_stats = []
        se_stats = []

        n_cohorts = len(cohorts_checked)
        for i in range(n_cohorts):
            for j in range(i + 1, n_cohorts):
                (fst, se) = self.average_fst(
                    region=region,
                    cohort1_query=cohort_queries[i],
                    cohort2_query=cohort_queries[j],
                    sample_sets=sample_sets,
                    cohort_size=cohort_size,
                    min_cohort_size=min_cohort_size,
                    max_cohort_size=max_cohort_size,
                    n_jack=n_jack,
                    site_mask=site_mask,
                    site_class=site_class,
                    random_seed=random_seed,
                )
                cohort1_ids.append(cohort_ids[i])
                cohort2_ids.append(cohort_ids[j])
                fst_stats.append(fst)
                se_stats.append(se)

        fst_df = pd.DataFrame(
            {
                "cohort1": cohort1_ids,
                "cohort2": cohort2_ids,
                "fst": fst_stats,
                "se": se_stats,
            }
        )

        return fst_df

    @check_types
    @doc(
        summary="""
            Plot a heatmap of pairwise average Fst values.
        """,
        parameters=dict(
            annotate_se="If True, show standard error values in the upper triangle of the plot.",
            kwargs="Passed through to `px.imshow()`",
        ),
    )
    def plot_pairwise_average_fst(
        self,
        fst_df: fst_params.df_pairwise_fst,
        annotation: fst_params.annotation = None,
        zmin: Optional[plotly_params.zmin] = 0.0,
        zmax: Optional[plotly_params.zmax] = None,
        text_auto: plotly_params.text_auto = ".3f",
        color_continuous_scale: plotly_params.color_continuous_scale = "gray_r",
        width: plotly_params.fig_width = None,
        height: plotly_params.fig_height = None,
        row_height: plotly_params.height = 40,
        col_width: plotly_params.width = 50,
        show: plotly_params.show = True,
        renderer: plotly_params.renderer = None,
        **kwargs,
    ):
        # Obtain a list of all cohorts analysed. N.B., preserve the order in
        # which the cohorts are provided in the input dataframe.
        cohorts = pd.unique(fst_df[["cohort1", "cohort2"]].values.flatten())

        # Create a dataframe to visualise as a heatmap.
        fig_df = pd.DataFrame(columns=cohorts, index=cohorts)

        # Set up plot title.
        title = "<i>F</i><sub>ST</sub>"
        if annotation is not None:
            title += " ⧅ " + annotation

        # Fill the figure dataframe from the Fst dataframe.
        for cohort1, cohort2, fst, se in fst_df.itertuples(index=False):
            fig_df.loc[cohort2, cohort1] = fst
            if annotation == "standard error":
                fig_df.loc[cohort1, cohort2] = se
            elif annotation == "Z score":
                zs = fst / se
                fig_df.loc[cohort1, cohort2] = zs
            else:
                fig_df.loc[cohort1, cohort2] = fst

        # Don't colour the plot if the upper triangle is SE or Z score,
        # as the colouring doesn't really make sense.
        if annotation is not None and zmax is None:
            zmax = 1e9

        # Dynamically size the figure based on number of cohorts.
        if height is None:
            height = 100 + len(cohorts) * row_height
        if width is None:
            width = 200 + len(cohorts) * col_width

        # Create plot.
        with np.errstate(invalid="ignore"):
            fig = px.imshow(
                img=fig_df,
                zmin=zmin,
                zmax=zmax,
                width=width,
                height=height,
                text_auto=text_auto,
                color_continuous_scale=color_continuous_scale,
                aspect="auto",
                **kwargs,
            )
        fig.update_layout(
            plot_bgcolor="rgba(0,0,0,0)",
            coloraxis_showscale=False,
            title=title,
        )
        fig.update_yaxes(showgrid=False, linecolor="black")
        fig.update_xaxes(showgrid=False, linecolor="black")

        if show:  # pragma: no cover
            fig.show(renderer=renderer)
            return None
        else:
            return fig



import collections
import operator

from Bio.Seq import Seq  # type: ignore

VariantEffect = collections.namedtuple(
    "VariantEffect",
    (
        "effect",
        "impact",
        "chrom",
        "pos",
        "ref",
        "alt",
        "vlen",
        "ref_start",
        "ref_stop",
        "strand",
        "ref_codon",
        "alt_codon",
        "codon_change",
        "aa_pos",
        "ref_aa",
        "alt_aa",
        "aa_change",
    ),
)
null_effect = VariantEffect(*([None] * len(VariantEffect._fields)))


class Annotator(object):
    def __init__(self, genome, genome_features):
        """
        An annotator.

        Parameters
        ----------
        genome : zarr hierarchy
            Reference genome.
        genome_features : pandas dataframe
            Dataframe with genome annotations.

        """

        # store initialisation parameters
        self._genome = genome
        self._genome_cache = dict()
        self._genome_features_cache = None

        genome_features = genome_features[
            (genome_features.end - genome_features.start) > 0
        ]
        self._genome_features_cache = genome_features

        # index features by ID
        self._idx_feature_id = self._genome_features_cache.set_index("ID")

        # index features by parent ID
        self._idx_parent_id = self._genome_features_cache.set_index("Parent")

    def get_feature(self, feature_id):
        return self._idx_feature_id.loc[feature_id]

    def get_children(self, feature_id):
        return self._idx_parent_id.loc[feature_id]

    def get_ref_seq(self, chrom, start, stop):
        """Accepts 1-based coords."""
        try:
            seq = self._genome_cache[chrom]
        except KeyError:
            seq = self._genome[chrom][:]
            self._genome_cache[chrom] = seq
        ref_seq = seq[start - 1 : stop]
        ref_seq = ref_seq.tobytes().decode()
        return ref_seq

    def get_ref_allele_coords(self, chrom, pos, ref):
        # N.B., use one-based inclusive coordinate system (like GFF3) throughout
        ref_start = pos
        ref_stop = pos + len(ref) - 1

        # check the reference allele matches the reference sequence
        ref_seq = self.get_ref_seq(chrom, ref_start, ref_stop).lower()
        assert ref_seq == ref.lower(), (
            "reference allele does not match reference sequence, "
            "expected %r, found %r" % (ref_seq, ref.lower())
        )

        return ref_start, ref_stop

    def get_effects(self, transcript, variants, progress=None):
        children = self.get_children(transcript).sort_values("start")
        feature = self.get_feature(transcript)

        # make sure all alleles are uppercase
        variants.ref_allele = variants.ref_allele.str.upper()
        variants.alt_allele = variants.alt_allele.str.upper()

        # get transcript children
        cdss = list(children[children.type == "CDS"].itertuples())
        exons = list(children[children.type == "exon"].itertuples())
        utr5 = list(children[children.type == "five_prime_UTR"].itertuples())
        utr3 = list(children[children.type == "three_prime_UTR"].itertuples())
        introns = [(x.end + 1, y.start - 1) for x, y in zip(exons[:-1], exons[1:])]

        effect_values = []
        impact_values = []
        ref_codon_values = []
        alt_codon_values = []
        aa_pos_values = []
        ref_aa_values = []
        alt_aa_values = []
        aa_change_values = []

        feature_contig = feature.contig
        feature_start = feature.start
        feature_stop = feature.end
        feature_strand = feature.strand

        variant_iterator = variants.itertuples(index=True)
        if progress:
            variant_iterator = progress(
                variant_iterator, desc="Compute SNP effects", total=len(variants)
            )

        for row in variant_iterator:
            # some parameters
            chrom = feature_contig
            pos = row.position
            ref = row.ref_allele
            alt = row.alt_allele

            # obtain start and stop coordinates of the reference allele
            ref_start, ref_stop = self.get_ref_allele_coords(chrom, pos, ref)

            # set up the common effect parameters
            base_effect = null_effect._replace(
                chrom=chrom,
                pos=pos,
                ref=ref,
                alt=alt,
                vlen=len(alt) - len(ref),
                ref_start=ref_start,
                ref_stop=ref_stop,
                strand=feature_strand,
            )

            # reference allele falls within current transcript
            assert feature_start <= ref_start <= ref_stop <= feature_stop

            effect = _get_within_transcript_effect(
                ann=self,
                base_effect=base_effect,
                cdss=cdss,
                utr5=utr5,
                utr3=utr3,
                introns=introns,
            )

            effect_values.append(effect.effect)
            impact_values.append(effect.impact)
            ref_codon_values.append(effect.ref_codon)
            alt_codon_values.append(effect.alt_codon)
            aa_pos_values.append(effect.aa_pos)
            ref_aa_values.append(effect.ref_aa)
            alt_aa_values.append(effect.alt_aa)
            aa_change_values.append(effect.aa_change)

        variants["transcript"] = transcript
        variants["effect"] = effect_values
        variants["impact"] = impact_values
        variants["ref_codon"] = ref_codon_values
        variants["alt_codon"] = alt_codon_values
        variants["aa_pos"] = aa_pos_values
        variants["ref_aa"] = ref_aa_values
        variants["alt_aa"] = alt_aa_values
        variants["aa_change"] = aa_change_values

        return variants


def _get_within_transcript_effect(ann, base_effect, cdss, utr5, utr3, introns):
    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop

    # find coding sequence that overlaps the reference allele
    within_cdss = [
        cds for cds in cdss if cds.start <= ref_start and cds.end >= ref_stop
    ]
    if within_cdss:
        return _get_within_cds_effect(ann, base_effect, within_cdss[0], cdss)

    within_introns = [
        (start, stop)
        for (start, stop) in introns
        if start <= ref_start and stop >= ref_stop
    ]
    if within_introns:
        return _get_within_intron_effect(
            base_effect=base_effect, intron=within_introns[0]
        )

    within_utr5 = [x for x in utr5 if x.start <= ref_start and x.end >= ref_stop]
    if within_utr5:
        effect = base_effect._replace(effect="FIVE_PRIME_UTR", impact="LOW")
        return effect

    within_utr3 = [x for x in utr3 if x.start <= ref_start and x.end >= ref_stop]
    if within_utr3:
        effect = base_effect._replace(effect="THREE_PRIME_UTR", impact="LOW")
        return effect

    # If none of the above, all we can say is that the variant hits
    # a transcript.
    effect = base_effect._replace(effect="TRANSCRIPT", impact="MODIFIER")
    return effect


def _get_cds_effect(ann, base_effect, cds, cdss):
    # setup common effect parameters
    base_effect = base_effect._replace(
        cds_id=cds.ID,
        cds_start=cds.start,
        cds_stop=cds.end,
        cds_strand=cds.strand,
    )

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    cds_start = cds.start
    cds_stop = cds.end

    # reference allele falls within current transcript
    assert cds_start <= ref_start <= ref_stop <= cds_stop
    return _get_within_cds_effect(ann, base_effect, cds, cdss)


def _get_within_cds_effect(ann, base_effect, cds, cdss):
    # convenience
    chrom = base_effect.chrom
    pos = base_effect.pos
    ref = base_effect.ref
    alt = base_effect.alt
    strand = base_effect.strand

    # obtain amino acid change
    (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
        aa_pos,
        ref_aa,
        alt_aa,
    ) = _get_aa_change(ann, chrom, pos, ref, alt, cds, cdss)

    # setup common effect parameters
    base_effect = base_effect._replace(
        ref_codon=ref_codon,
        alt_codon=alt_codon,
        codon_change="%s/%s" % (ref_codon, alt_codon),
        aa_pos=aa_pos,
        ref_aa=ref_aa,
        alt_aa=alt_aa,
        aa_change="%s%s%s" % (ref_aa, aa_pos, alt_aa),
    )

    if len(ref) == 1 and len(alt) == 1:
        # SNPs

        if ref_aa == alt_aa:
            # TODO SYNONYMOUS_START and SYNONYMOUS_STOP

            # variant causes a codon that produces the same amino acid
            # e.g.: Ttg/Ctg, L/L
            effect = base_effect._replace(effect="SYNONYMOUS_CODING", impact="LOW")

        elif ref_aa == "M" and ref_cds_start == 0:
            # variant causes start codon to be mutated into a non-start codon.
            # e.g.: aTg/aGg, M/R
            effect = base_effect._replace(effect="START_LOST", impact="HIGH")

        elif ref_aa == "*":
            # variant causes stop codon to be mutated into a non-stop codon
            # e.g.: Tga/Cga, */R
            effect = base_effect._replace(effect="STOP_LOST", impact="HIGH")

        elif alt_aa == "*":
            # variant causes a STOP codon e.g.: Cag/Tag, Q/*
            effect = base_effect._replace(effect="STOP_GAINED", impact="HIGH")

        else:
            # TODO NON_SYNONYMOUS_START and NON_SYNONYMOUS_STOP

            # variant causes a codon that produces a different amino acid
            # e.g.: Tgg/Cgg, W/R
            effect = base_effect._replace(
                effect="NON_SYNONYMOUS_CODING", impact="MODERATE"
            )

    else:
        # INDELs and MNPs

        if (len(alt) - len(ref)) % 3:
            # N.B., this case covers both simple INDELs and complex
            # polymorphisms

            # insertion or deletion causes a frame shift
            # e.g.: An indel size is not multple of 3
            effect = base_effect._replace(effect="FRAME_SHIFT", impact="HIGH")

        elif len(ref) == 1 and len(ref) < len(alt):
            # simple insertions

            # figure out if there has been a codon change or not
            is_codon_changed = (strand == "+" and ref_aa[0] != alt_aa[0]) or (
                strand == "-" and ref_aa[-1] != alt_aa[-1]
            )

            if is_codon_changed:
                # one codon is changed and one or many codons are inserted
                # e.g.: An insert of size multiple of three, not at codon
                # boundary
                effect = base_effect._replace(
                    effect="CODON_CHANGE_PLUS_CODON_INSERTION", impact="MODERATE"
                )

            else:
                # one or many codons are inserted
                # e.g.: An insert multiple of three in a codon boundary
                effect = base_effect._replace(
                    effect="CODON_INSERTION", impact="MODERATE"
                )

        elif len(alt) == 1 and len(ref) > len(alt):
            # simple deletions

            # figure out if there has been a codon change or not
            is_codon_changed = (strand == "+" and ref_aa[0] != alt_aa[0]) or (
                strand == "-" and ref_aa[-1] != alt_aa[-1]
            )

            if is_codon_changed:
                # one codon is changed and one or many codons are deleted
                # e.g.: A deletion of size multiple of three, not at codon
                # boundary
                effect = base_effect._replace(
                    effect="CODON_CHANGE_PLUS_CODON_DELETION", impact="MODERATE"
                )

            else:
                # one or many codons are deleted
                # e.g.: A deletions multiple of three in a codon boundary
                effect = base_effect._replace(
                    effect="CODON_DELETION", impact="MODERATE"
                )

        elif len(ref) == len(alt):
            # MNPs
            effect = base_effect._replace(effect="CODON_CHANGE", impact="MODERATE")

        else:
            # TODO in-frame complex variation (MNP + INDEL)
            effect = base_effect._replace(
                effect="TODO in-frame complex variation (MNP + INDEL)", impact="UNKNOWN"
            )

    return effect


def _get_aa_change(ann, chrom, pos, ref, alt, cds, cdss):
    # obtain codon change
    (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
    ) = _get_codon_change(ann, chrom, pos, ref, alt, cds, cdss)

    # translate codon change to amino acid change
    ref_aa = str(Seq(ref_codon).translate())
    alt_aa = str(Seq(alt_codon).translate())
    aa_pos = (ref_cds_start // 3) + 1

    return (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
        aa_pos,
        ref_aa,
        alt_aa,
    )


def _get_codon_change(ann, chrom, pos, ref, alt, cds, cdss):
    # obtain reference allele coords relative to coding sequence
    ref_start, ref_stop = ann.get_ref_allele_coords(chrom, pos, ref)
    ref_cds_start, ref_cds_stop = _get_coding_position(ref_start, ref_stop, cds, cdss)

    # calculate position of reference allele start within codon
    ref_start_phase = ref_cds_start % 3

    if cds.strand == "+":
        # obtain any previous nucleotides to complete the first codon
        prefix = ann.get_ref_seq(
            chrom=chrom, start=ref_start - ref_start_phase, stop=ref_start - 1
        ).lower()

        # begin constructing reference and alternate codon sequences
        ref_codon = prefix + ref
        alt_codon = prefix + alt

        # obtain any subsequence nucleotides to complete the last codon
        if len(ref_codon) % 3:
            ref_stop_phase = len(ref_codon) % 3
            suffix = ann.get_ref_seq(
                chrom=chrom, start=ref_stop + 1, stop=ref_stop + 3 - ref_stop_phase
            )
            suffix = str(suffix).lower()
            ref_codon += suffix

        if len(alt_codon) % 3:
            alt_stop_phase = len(alt_codon) % 3
            suffix = ann.get_ref_seq(
                chrom=chrom, start=ref_stop + 1, stop=ref_stop + 3 - alt_stop_phase
            ).lower()
            alt_codon += suffix

    else:
        # N.B., we are on the reverse strand, so position reported for
        # variant is actually position at the *end* of the reference allele
        # which is particularly important for deletions

        # we will construct everything for the forward strand (i.e., back-to-
        # front) then take reverse complement afterwards at the end of this
        # code block

        # obtain any previous nucleotides to complete the first codon
        prefix = ann.get_ref_seq(
            chrom=chrom, start=ref_stop + 1, stop=ref_stop + ref_start_phase
        ).lower()

        # begin constructing reference and alternate codon sequences
        ref_codon = ref + prefix
        alt_codon = alt + prefix

        # obtain any subsequence nucleotides to complete the last codon
        if len(ref_codon) % 3:
            ref_stop_phase = len(ref_codon) % 3
            suffix = ann.get_ref_seq(
                chrom=chrom, start=ref_start - 3 + ref_stop_phase, stop=ref_start - 1
            ).lower()
            ref_codon = suffix + ref_codon

        if len(alt_codon) % 3:
            alt_stop_phase = len(alt_codon) % 3
            suffix = ann.get_ref_seq(
                chrom=chrom, start=ref_start - 3 + alt_stop_phase, stop=ref_start - 1
            ).lower()
            alt_codon = suffix + alt_codon

        # take reverse complement
        ref_codon = str(Seq(ref_codon).reverse_complement())
        alt_codon = str(Seq(alt_codon).reverse_complement())

    return ref_cds_start, ref_cds_stop, ref_start_phase, ref_codon, alt_codon


def _get_coding_position(ref_start, ref_stop, cds, cdss):
    if cds.strand == "+":
        # sort exons
        cdss = sorted(cdss, key=operator.attrgetter("start"))

        # find index of overlapping exons in all exons
        cds_index = [f.start for f in cdss].index(cds.start)

        # find offset
        offset = sum([f.end - f.start + 1 for f in cdss[:cds_index]])

        # find ref cds position
        ref_cds_start = offset + (ref_start - cds.start)
        ref_cds_stop = offset + (ref_stop - cds.start)

    else:
        # sort exons (backwards this time)
        cdss = sorted(cdss, key=operator.attrgetter("end"), reverse=True)

        # find index of overlapping exons in all exons
        cds_index = [f.end for f in cdss].index(cds.end)

        # find offset
        offset = sum([cds.end - cds.start + 1 for cds in cdss[:cds_index]])

        # find ref cds position
        ref_cds_start = offset + (cds.end - ref_stop)
        ref_cds_stop = offset + (cds.end - ref_start)

    return ref_cds_start, ref_cds_stop


def _get_within_intron_effect(base_effect, intron):
    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    ref = base_effect.ref
    alt = base_effect.alt
    intron_start, intron_stop = intron
    strand = base_effect.strand
    if strand == "+":
        intron_5prime_dist = ref_start - (intron_start - 1)
        intron_3prime_dist = ref_stop - (intron_stop + 1)

    else:
        intron_5prime_dist = (intron_stop + 1) - ref_stop
        intron_3prime_dist = (intron_start - 1) - ref_start

    intron_min_dist = min(intron_5prime_dist, -intron_3prime_dist)

    if len(ref) == 1 and len(alt) == 1:
        # SNPs

        if intron_min_dist <= 2:
            # splice site variation
            effect = base_effect._replace(effect="SPLICE_CORE", impact="HIGH")

        elif intron_min_dist <= 7:
            # splice site variation
            effect = base_effect._replace(effect="SPLICE_REGION", impact="MODERATE")

        else:
            # intron modifier
            effect = base_effect._replace(effect="INTRONIC", impact="MODIFIER")

    else:
        # TODO intronic INDELs and MNPs
        effect = base_effect._replace(effect="TODO intronic indels and MNPs")

    return effect
